/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.	Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.
   
   DRAW.C - Editor routines.
   */

/* the includes */
#include <math.h>
#include "deu.h"
#include "levels.h"

void DrawMapLinesQuick();
void DrawInfoBar(BCINT, BCINT);

#if defined(GAME_HEXEN)
	int tff = TF_EASY | TF_MEDIUM | TF_HARD | TF_FIGHTER | TF_CLERIC | TF_MAGE | TF_SINGLE | TF_COOP | TF_DEATHM;
#else
	int tff = TF_EASY | TF_MEDIUM | TF_HARD | TF_SINGLE | TF_COOP | TF_DEATHM;
#endif

Bool DrawThingCircles;

void DrawMapQuick(void)
{
	BCINT clip, yoff;

	clip = (InfoShown ? 13 : 0);
	yoff = 17;
	
	if (!BrowseMap)
	{
		GrSetClipBox (0, yoff, ScrMaxX, ScrMaxY - clip);
		clearviewport();
		DrawMapLinesQuick();
		GrResetClipBox ();
	}
	else
	{
		DrawMapLinesQuick();
		DrawScreenBox3D( 0, ScrMaxY - 12, ScrMaxX, ScrMaxY);
		SetColor( WHITE);
		DrawScreenText( ScrMaxX - 450, ScrMaxY - 9, "Browse: Up,Down,PgUp,PgDn,Home,End  View: Q,T,V,L,S,-,+");
		DrawScreenText( 5, ScrMaxY - 9, "Viewing:");
		if (Level->wadfile->type[0] == 'I')
			SetColor (YELLOW);
		DrawScreenText (5+9*8, ScrMaxY - 9, Level->dir.name);
		if (BehaviorSize == 0)
		{
			SetColor (YELLOW);
			DrawScreenText (5,	5, "This map is not in the Hexen format, so all its specials will be zeroed");
			DrawScreenText (5, 15, "if you edit it. Use ZWADCONV to convert the map if you want to keep them.");
		}
	}
}

void DrawMapLinesQuick()
{
	BCINT n;

	SetColor(LIGHTGRAY);
	for (n = 0; n < NumLineDefs; n++)
	{
		DrawMapLine
		(
			Vertexes[LineDefs[n].start].x,
			Vertexes[LineDefs[n].start].y,
			Vertexes[LineDefs[n].end].x,
			Vertexes[LineDefs[n].end].y
		);
	}

	/* draw in the things */
	SetColor(DARKGRAY);
	SetLineDotted (TRUE);
	for (n = 0; n < NumThings; n++)
	{
		DrawMapLine
		(
			Things[n].xpos - (OBJSIZE * 2),
			Things[n].ypos,
			Things[n].xpos + (OBJSIZE * 2),
			Things[n].ypos
		);
		DrawMapLine
		(
			Things[n].xpos,
			Things[n].ypos - (OBJSIZE * 2),
			Things[n].xpos,
			Things[n].ypos + (OBJSIZE * 2)
		);
	}
	SetLineDotted (FALSE);
}

void DrawThingAngle (int x, int y, int m, int angle, int style)
{
	int pt[3][2];

	switch (angle)
	{
	case 0:
		pt[1][0] = x + m;
		pt[1][1] = y;

		m >>= 1;
		pt[0][0] = pt[2][0] = x;
		pt[0][1] = y + m;
		pt[2][1] = y - m;
		break;
	case 180:
		pt[1][0] = x - m;
		pt[1][1] = y;

		m >>= 1;
		pt[0][0] = pt[2][0] = x;
		pt[0][1] = y + m;
		pt[2][1] = y - m;
		break;
	case 90:
		pt[1][0] = x;
		pt[1][1] = y + m;

		m >>= 1;
		pt[0][1] = pt[2][1] = y;
		pt[0][0] = x + m;
		pt[2][0] = x - m;
		break;
	case 270:
		pt[1][0] = x;
		pt[1][1] = y - m;

		m >>= 1;
		pt[0][1] = pt[2][1] = y;
		pt[0][0] = x + m;
		pt[2][0] = x - m;
		break;
	case 45:
		m = (m * 181) >> 8;
		pt[1][0] = x + m;
		pt[1][1] = y + m;

		m >>= 1;
		pt[0][0] = x + m;
		pt[0][1] = y - m;
		pt[2][0] = x - m;
		pt[2][1] = y + m;
		break;
	case 315:
		m = (m * 181) >> 8;
		pt[1][0] = x + m;
		pt[1][1] = y - m;

		m >>= 1;
		pt[0][0] = x + m;
		pt[0][1] = y + m;
		pt[2][0] = x - m;
		pt[2][1] = y - m;
		break;
	case 225:
		m = (m * 181) >> 8;
		pt[1][0] = x - m;
		pt[1][1] = y - m;

		m >>= 1;
		pt[0][0] = x + m;
		pt[0][1] = y - m;
		pt[2][0] = x - m;
		pt[2][1] = y + m;
		break;
	case 135:
		m = (m * 181) >> 8;
		pt[1][0] = x - m;
		pt[1][1] = y + m;

		m >>= 1;
		pt[0][0] = x + m;
		pt[0][1] = y + m;
		pt[2][0] = x - m;
		pt[2][1] = y - m;
		break;
	default:
		{
			double r, ang;
			int cr, sr;

			r = (double)m;
			ang = (double)angle * M_PI / 180.0;
			cr = r * cos (ang);
			sr = r * sin (ang);

			pt[1][0] = x + cr;
			pt[1][1] = y + sr;
			cr >>= 1;
			sr >>= 1;
			pt[0][0] = x - sr;
			pt[0][1] = y + cr;
			pt[2][0] = x + sr;
			pt[2][1] = y - cr;
		}
		break;
	}
	if (style)
	{
		DrawMapVector (x, y, pt[1][0], pt[1][1]);
	}
	else
	{
		int i;
		for (i = 0; i < 3; ++i)
		{
			pt[i][0] = SCREENX(pt[i][0]);
			pt[i][1] = SCREENY(pt[i][1]);
		}
		GrFilledConvexPolygon (3, pt, 0);
	}
}

/*
   draw the actual game map 						
   */

void DrawMap( BCINT editmode, BCINT gsize, Bool drawgrid) /* SWAP! */
{
	int val, n, m, o, max;
	BCINT sd1floor, sd2floor, sd1ceil, sd2ceil;
	BCINT width, height;
	Bool NotDrawn;

	max = (InfoShown ? 13 : 0);

	ClearMapScreen(17, max);

	/* draw the grid */
	if (drawgrid && gsize > 0)
	{
		int grid = gsize;
		int gridmask = ~(grid - 1);
		int mapx0 = (int)(MAPX(0)) & gridmask;
		int mapx1 = (int)(MAPX(ScrMaxX) + grid) & gridmask;
		int mapy1 = (int)(MAPY(0)+grid) & gridmask;
		int mapy0 = (int)(MAPY(ScrMaxY - max) - grid) & gridmask;

		/* HideMousePointer(); */

		// [RH] Do not try to draw the grid beyond the editable area.
		if (mapx0 < -32768)
			mapx0 = -32768;
		else if (mapx0 > 32768)
			mapx0 = 32768;
		if (mapy0 < -32768)
			mapy0 = -32768;
		else if (mapy0 > 32768)
			mapy0 = 32768;
		if (mapx1 < -32768)
			mapx1 = -32768;
		else if (mapx1 > 32768)
			mapx1 = 32768;
		if (mapy1 < -32768)
			mapy1 = -32768;
		else if (mapy1 > 32768)
			mapy1 = 32768;

		if
		(
			!GridAccurate && Scale < 8.0 &&
			(
				((SCREENX(mapx0 + grid) - SCREENX(mapx0)) < 8) ||
				((SCREENY(mapy0) - SCREENY(mapy0 + grid)) < 9)
			)
		)
		{
			while (SCREENX(mapx0 + grid) - SCREENX(mapx0) < 8)
			{
				grid *= 2;
			}
		}
		SetColor(GRIDCOLOR);
		if (GridDashed)
			SetLineDotted (TRUE);
		for (n = mapx0; n <= mapx1; n += grid)
			DrawMapLine(n, mapy0, n, mapy1);
		for (n = mapy0; n <= mapy1; n += grid)
			DrawMapLine(mapx0, n, mapx1, n);
		SetLineDotted (FALSE);
		if (FloorGridShown && Scale > 0.07)
		{
			SetColor(FLOORGRIDCOLOR);
			for (o = (mapy0+63)&~63; o <= mapy1; o += 64)
			{
				for (n = (mapx0+63)&~63; n <= mapx1; n += 64)
				{
					DrawMapLine (n-5, o, n+5, o);
					DrawMapLine (n, o-5, n, o+5);
				}
			}
		}
	}
	
	/* draw the linedefs to form the map */
	switch (editmode)
	{
	case OBJ_THINGS:
		for (n = 0; n < NumLineDefs; n++)
		{
			if (LineDefs[ n].flags & 1 || LineDefs[n].sidedef2 == -1)
				SetColor(LIGHTGRAY);
			else if (LineDefs[ n].flags & 2)
				SetColor(LINEDEFNOPASS);
			else if (LineDefs[ n].flags & 32)
				SetColor(LINEDEFSECRET);
			else if (LineDefs[ n].flags & 64)
				SetColor(LINEDEFNOSOUND);
			else if (LineDefs[ n].flags & 128)
				SetColor(LINEDEFNOMAP);
			else
				SetColor(DARKGRAY);
			DrawMapLine(Vertexes[ LineDefs[ n].start].x, Vertexes[ LineDefs[ n].start].y,
						Vertexes[ LineDefs[ n].end].x, Vertexes[ LineDefs[ n].end].y);
		}
		break;

	case OBJ_VERTEXES:
		SetColor (DARKGRAY);
		SetTextCentered (TRUE);
		SetColor (WHITE);
		SetTextBg (GRAY);
		for (n = 0; n < NumLineDefs; n++)
		{
			DrawMapVector
			(
				Vertexes[LineDefs[n].start].x,
				Vertexes[LineDefs[n].start].y,
				Vertexes[LineDefs[n].end].x,
				Vertexes[LineDefs[n].end].y
			);
			if (ShowDistance)
			{
				/* HideMousePointer(); */
				o = (Vertexes[ LineDefs[ n].start].x + Vertexes[ LineDefs[ n].end].x) / 2;
				m = (Vertexes[ LineDefs[ n].start].y + Vertexes[ LineDefs[ n].end].y) / 2;
				val = ComputeDist
					  (
						  Vertexes[LineDefs[n].end].x - Vertexes[LineDefs[n].start].x,
						  Vertexes[LineDefs[n].end].y - Vertexes[LineDefs[n].start].y
					  );
				DrawScreenTextSmall(SCREENX(o), SCREENY(m) - 6, " %d ", val);
			}
		}
		SetColor (DARKGRAY);
		SetTextBg (-1);
		SetTextCentered (FALSE);
		/* ShowMousePointer(); */
		break;

	case OBJ_LINEDEFS:
		if (!AutoMap)
		{
			for (n = 0; n < NumLineDefs; n++)
			{
				if (LineDefs[ n].special > 0)
				{
					char *text = GetLineDefArgs (LineDefs[n].special);
					int k;

					for (k = 0; k < 5; ++k)
					{
						if ((text[k] == '$' || text[k] == '%') && LineDefs[n].arg[k] != 0)
						{
							break;
						}
					}
					if (k < 5)
						SetColor( LINEDEFTAGGED);
					else
						SetColor( LIGHTGREEN);
				}
				else if (LineDefs[ n].arg[0] > 0)
					SetColor( LINEDEFTAGGED);
				else if (LineDefs[ n].flags & 1 || LineDefs[n].sidedef2 == -1)
					SetColor(LIGHTGRAY); 
				else if (LineDefs[ n].flags & 2)
					SetColor(LINEDEFNOPASS);
				else if (LineDefs[ n].flags & 32)
					SetColor(LINEDEFSECRET);
				else if (LineDefs[ n].flags & 64)
					SetColor(LINEDEFNOSOUND);
				else if (LineDefs[ n].flags & 128)
					SetColor(LINEDEFNOMAP);
				else
					SetColor(DARKGRAY);
				o = (Vertexes[ LineDefs[ n].start].x + Vertexes[ LineDefs[ n].end].x) / 2;
				m = (Vertexes[ LineDefs[ n].start].y + Vertexes[ LineDefs[ n].end].y) / 2;
				DrawMapLine( Vertexes[ LineDefs[ n].start].x, Vertexes[ LineDefs[ n].start].y,
					Vertexes[ LineDefs[ n].end].x, Vertexes[ LineDefs[ n].end].y);
			}
		}
		else
		{
			for (n = 0; n < NumLineDefs; n++)
			{
				sd1floor = Sectors[SideDefs[LineDefs[ n].sidedef1].sector].floorh;
				sd2floor = Sectors[SideDefs[LineDefs[ n].sidedef2].sector].floorh;
				sd1ceil = Sectors[SideDefs[LineDefs[ n].sidedef1].sector].ceilh;
				sd2ceil = Sectors[SideDefs[LineDefs[ n].sidedef2].sector].ceilh;

				SetColor(RED);
				NotDrawn = FALSE;
				if (sd1floor == sd2floor)
					NotDrawn = TRUE;
				if (sd1floor > sd2floor)
				{
					NotDrawn = FALSE;
					SetColor(BROWN);
				}
				if (sd2floor > sd1floor)
				{
					NotDrawn = FALSE;
					SetColor(BROWN);
				}
				if (sd1ceil < sd2ceil)
				{
					NotDrawn = FALSE;
					SetColor(YELLOW);
				}
				if (sd2ceil < sd1ceil)
				{
					NotDrawn = FALSE;
					SetColor(YELLOW);
				}
				if ((LineDefs[ n].flags & 32769))	/* IMPASSIBLE | totally blocking */
				{
					NotDrawn = FALSE;
					SetColor(RED);
				}
				if (LineDefs[ n].flags & 32)
				{ /* SECRET */
					NotDrawn = FALSE;
					SetColor(RED);
				}
				if (LineDefs[ n].flags & 128)  /* NOT ON MAP */
				{
					NotDrawn = TRUE;
				}
				if (LineDefs[ n].flags & 256)
				{ /* ALREADY ON MAP */
					NotDrawn = FALSE;
					SetColor(RED);
				}

				if (!NotDrawn)
				{
					o = (Vertexes[ LineDefs[ n].start].x + Vertexes[ LineDefs[ n].end].x) / 2;
					m = (Vertexes[ LineDefs[ n].start].y + Vertexes[ LineDefs[ n].end].y) / 2;
					DrawMapLine
					(
						Vertexes[LineDefs[n].start].x,
						Vertexes[LineDefs[n].start].y,
						Vertexes[LineDefs[n].end].x,
						Vertexes[LineDefs[n].end].y
					);
				}
			}
		}
		break;

	case OBJ_SECTORS:
		for (n = 0; n < NumLineDefs; n++)
		{
			if ((m = LineDefs[ n].sidedef1) < 0 || (m = SideDefs[ m].sector) < 0)
			{
				SetColor( CYAN);
			}
			else
			{
				if (Sectors[ m].tag > 0)
					SetColor( SECTORTAGGED);
				else if (Sectors[ m].special == 26 || Sectors[ m].special == 27)
					SetColor( SECTORPAINFUL);
				else if (Sectors[ m].special == 200)
					SetColor( SECTORSECRET);
				else if (Sectors[ m].special == 1 || Sectors[ m].special == 2 || Sectors[ m].special == 3 || Sectors[ m].special == 4 || Sectors[ m].special == 8 || Sectors[ m].special == 12 || Sectors[ m].special == 17 || Sectors[ m].special == 199 || Sectors[ m].special == 198) 
					SetColor( SECTORLIGHT);
				else if (LineDefs[ n].flags & 1 || LineDefs[n].sidedef2 == -1)
					SetColor(LIGHTGRAY);
				else if (LineDefs[ n].flags & 2)
					SetColor(LINEDEFNOPASS);
				else if (LineDefs[ n].flags & 32)
					SetColor(LINEDEFSECRET);
				else if (LineDefs[ n].flags & 64)
					SetColor(LINEDEFNOSOUND);
				else if (LineDefs[ n].flags & 128)
					SetColor(LINEDEFNOMAP);
				else
					SetColor(DARKGRAY);

				if ((m = LineDefs[ n].sidedef2) >= 0)
				{
					if ((m = SideDefs[ m].sector) < 0)
						SetColor( LIGHTBLUE);
					else if (Sectors[ m].tag > 0)
						SetColor( SECTORTAGGED);
					else if (Sectors[ m].special == 7 || Sectors[ m].special == 5 || Sectors[ m].special == 16 || Sectors[ m].special == 4 || Sectors[ m].special == 11 || Sectors[ m].special == 6)
						SetColor( SECTORPAINFUL);
					else if (Sectors[ m].special == 9)
						SetColor( SECTORSECRET);
					else if (Sectors[ m].special == 2 || Sectors[ m].special == 3 || Sectors[ m].special == 13 || Sectors[ m].special == 12 || Sectors[ m].special == 1 || Sectors[ m].special == 8 || Sectors[ m].special == 17) 
						SetColor( SECTORLIGHT);
				}
			}
			DrawMapLine( Vertexes[ LineDefs[ n].start].x, Vertexes[ LineDefs[ n].start].y,
						Vertexes[ LineDefs[ n].end].x, Vertexes[ LineDefs[ n].end].y);
			o = (Vertexes[ LineDefs[ n].start].x + Vertexes[ LineDefs[ n].end].x) / 2;
			m = (Vertexes[ LineDefs[ n].start].y + Vertexes[ LineDefs[ n].end].y) / 2;
		}
		break;
	}
	
	/* draw in the vertices */
	if (editmode == OBJ_VERTEXES)
	{
		SetColor( LIGHTGREEN);
		for (n = 0; n < NumVertexes; n++)
		{
			DrawMapHollowBox (Vertexes[n].x - OBJSIZE/4, Vertexes[n].y - OBJSIZE/4, Vertexes[n].x + OBJSIZE/4, Vertexes[n].y + OBJSIZE/4);
		}
	}
	
	/* draw in the things */
	if (editmode == OBJ_THINGS) 
	{
		long step;

		if (ShowThings && Scale >= 0.25)
		{
			step = (long)(65536.0 / Scale);
		}
		else
		{
			step = 0;
		}
		for (n = 0; n < NumThings; n++)
		{
			if(!(Things[n].flags & tff) && Things[n].flags != 0)
				continue;

			if (step != 0 && !(GetThingExtra (Things[n].type) & TX_NOPIC) && GetThingSprite(Things[n].type)[0])
			{
				if (bioskey(1))
				{
					SetLineDotted (FALSE);
					return;
				}
				SetLineDotted (TRUE);
				if (SCREENX(Things[n].xpos) >= 0 && SCREENY(Things[n].ypos) >= 0)
				{
					if (Scale > 1.0)
					{
						m = GetThingRadius (Things[n].type, Things[n].arg[0]);

						SetColor(DARKERGRAY);
						DrawMapLine( Things[ n].xpos - ( OBJSIZE * 2 ), Things[ n].ypos, Things[ n].xpos + ( OBJSIZE * 2 ), Things[ n].ypos);
						DrawMapLine( Things[ n].xpos, Things[ n].ypos - ( OBJSIZE * 2 ), Things[ n].xpos, Things[ n].ypos + ( OBJSIZE * 2 ));
						if (!(GetThingExtra (Things[n].type) & TX_NOANGLE))
						{	// some things don't have angles
							DrawThingAngle (Things[n].xpos, Things[n].ypos, m, Things[n].angle, 1);
						}
					}
					GetPicSize(GetThingSprite(Things[n].type), &width, &height);
					DisplayPicFrac (SCREENX(Things[n].xpos) - (width * 32768)/step,
									SCREENY(Things[n].ypos)*step/65536 - height/2,
									ScrMaxX, ScrMaxY, GetThingSprite(Things[n].type), step, step);
				}
			}
			else
			{
				if (bioskey (1))
				{
					return;
				}
				SetColor( GetThingColour(Things[n].type));
				m = GetThingRadius (Things[n].type, Things[n].arg[0]);
				if (DrawThingCircles)
				{
					DrawMapCircle( Things[ n].xpos, Things[ n].ypos, m);
				}
				else
				{
					DrawMapTBox (Things[n].xpos - m, Things[n].ypos - m, Things[n].xpos + m, Things[n].ypos + m);
					SetColor (BLACK);
				}
				m -= 2;
				if (!ThingAngle || (GetThingExtra (Things[n].type) & TX_NOANGLE))
				{
					if (DrawThingCircles)
					{
						DrawMapLine( Things[n].xpos - m, Things[n].ypos, Things[n].xpos + m, Things[n].ypos);
						DrawMapLine( Things[n].xpos, Things[n].ypos - m, Things[n].xpos, Things[n].ypos + m);
					}
				}
				else
				{
					DrawThingAngle (Things[n].xpos, Things[n].ypos, m, Things[n].angle, DrawThingCircles);
				}
			}
		}
		SetLineDotted (FALSE);

		// [RH] Now draw any paths
		if (ShowPaths)
		{
			for (n = 0; n < NumThings; ++n)
			{
				char *args = GetThingArgs (Things[n].type);
				int extra = GetThingExtra (Things[n].type);
				if (args)
				{
					if (extra & (TX_PATH | TX_TOPATH))
					{
						int pathtype;

						if (extra & TX_TOPATH)
							pathtype = GetThingPathType (Things[n].type);
						else
							pathtype = Things[n].type;
						SetColor ((extra & TX_TOPATH) ? THINGTOPATH : THINGPATH);
						// Normal case: Draw an arrow to the next node on the path
						for (m = 0; m < 5; ++m)
						{
							if (args[m] == '*')
							{
								int arg = Things[n].arg[m];
								int next;

								if (args[m+1] == '+')
									arg += 256*Things[n].arg[m+1];

								next = FindThingByTID (pathtype, arg);
								if (next >= 0)
								{
									DrawMapVector (Things[n].xpos, Things[n].ypos, Things[next].xpos, Things[next].ypos);
								}
								break;
							}
						}
					}
				}
				else if (extra & TX_TOPATH)
				{ // Draw a dotted line to the first matching thing on the path
					int next = FindThingByTID (GetThingPathType (Things[n].type), Things[n].tid);
					if (next >= 0)
					{
						SetColor (THINGTOPATH);
						SetLineDotted (TRUE);
						DrawMapLine (Things[n].xpos, Things[n].ypos, Things[next].xpos, Things[next].ypos);
						SetLineDotted (FALSE);
					}
				}
				else if (!(extra & TX_DEATHSCRIPT) && Things[n].special == 229 && Things[n].arg[0] == 0)
				{ // Special case for things with the Thing_SetGoal special: Draw an arrow to
				  // the specified node on the path.
					int next = FindThingByTID (9024, Things[n].arg[1]);
					if (next >= 0)
					{
						SetColor (THINGTOPATH);
						DrawMapVector (Things[n].xpos, Things[n].ypos, Things[next].xpos, Things[next].ypos);
					}
				}
			}
		}
	}
	else
	{
		if (!AutoMap)
		{
			for (n = 0; n < NumThings; n++)
			{
				SetColor(DARKGRAY);
				m = GetThingRadius(Things[n].type, Things[n].arg[0]);
				DrawMapLine( Things[ n].xpos - ( OBJSIZE * 2 ), Things[ n].ypos, Things[ n].xpos + ( OBJSIZE * 2 ), Things[ n].ypos);
				DrawMapLine( Things[ n].xpos, Things[ n].ypos - ( OBJSIZE * 2 ), Things[ n].xpos, Things[ n].ypos + ( OBJSIZE * 2 ));
			}
		}
	}

	if (BrowseMap)
	{
		DrawScreenBox3D( 0, ScrMaxY - 12, ScrMaxX, ScrMaxY);
		SetColor( WHITE);
		DrawScreenText( ScrMaxX - 450, ScrMaxY - 9, "Browse: Up,Down,PgUp,PgDn,Home,End  View: Q,T,V,L,S,-,+");
		DrawScreenText( 5, ScrMaxY - 9, "Viewing:");
		if (Level->wadfile->type[0] == 'I')
			SetColor (YELLOW);
		DrawScreenText (5+9*8, ScrMaxY - 9, Level->dir.name);
		if (BehaviorSize == 0)
		{
			SetColor (YELLOW);
			DrawScreenText (5,	5, "This map is not in the Hexen format, so all its specials will be zeroed");
			DrawScreenText (5, 15, "if you edit it. Use ZWADCONV to convert the map if you want to keep them.");
		}
	}
	else
	{
		/* draw in the title bar */
		DrawScreenBox3D( 0, 0, ScrMaxX, 16);
		SetColor( WHITE);
		DrawScreenText( 7,	4, "File Edit Search Mode Misc(F8) Objects(F9) Check(F10)");
		DrawScreenText( 7,	6, "_    _    _      _     _       _           _");
		DrawScreenText( ScrMaxX - 50, 4, "Help");
		DrawScreenText( ScrMaxX - 50, 6, "_"); 
		
	}
	DrawInfoBar(editmode, gsize);
}

void DrawInfoBar(BCINT editmode, BCINT grid)
{
	if (InfoShown && !BrowseMap)
	{
		DrawScreenBox3D(0, ScrMaxY - 12, ScrMaxX, ScrMaxY);
		if (!AltInfo)
		{
			if (MadeMapChanges)
				DrawScreenTextSmall(5, ScrMaxY - 12, "%s on %s #", GetEditModeName( editmode), Level->dir.name);
			else if (MadeChanges)
				DrawScreenTextSmall(5, ScrMaxY - 12, "%s on %s *", GetEditModeName( editmode), Level->dir.name);
			else
				DrawScreenTextSmall(5, ScrMaxY - 12, "%s on %s", GetEditModeName( editmode), Level->dir.name);
		}
		else
		{
			SetColor(BLUE);
			DrawScreenTextSmall(5, ScrMaxY - 12, "Undo: %s", BackupText);
		}
		SetColor(BLACK);
		/*
		if (Scale < 1.0)
			DrawScreenTextSmall(ScrMaxX - 140, ScrMaxY - 12, "Scale: 1/%d", (BCINT) (1.0 / Scale + 0.5));
		else
			DrawScreenTextSmall(ScrMaxX - 140, ScrMaxY - 12, "Scale: %d/1", (BCINT) Scale);
		*/
		DrawScreenTextSmall(ScrMaxX - 140, ScrMaxY - 12, "Scale: %.3f", Scale);
		DrawScreenTextSmall(ScrMaxX - 62, ScrMaxY - 12, "Grid: %d", grid);
		DrawScreenTextSmall(ScrMaxX - 360, ScrMaxY - 12, "Ctl: %d", CloseToLine);
		if (farcoreleft() < 50000L)
		{
			if (farcoreleft() < 20000L)
				SetColor(CYAN);
			else
				SetColor(RED);
		}
		DrawScreenTextSmall(ScrMaxX - 310, ScrMaxY - 12, "Mem: %lu", farcoreleft());
		SetColor(YELLOW);
		DrawScreenTextSmall(ScrMaxX - 205, ScrMaxY - 12, "Sel: %d", Count);
	}
}


/*
   center the map around the given coords
   */

void CenterMapAroundCoords( BCINT xpos, BCINT ypos)
{
	OrigX = xpos;
	OrigY = ypos;
	PointerX = ScrCenterX;
	PointerY = ScrCenterY;
}

/*
	center the map on screen and set the scale so it is all visible
	*/
void ShowAllMap (void)
{
	double yscale, xscale;
	CenterMapAroundCoords ((MapMinX + MapMaxX) / 2, (MapMinY + MapMaxY) / 2);
	yscale = (double)ScrMaxX / (double)(MapMaxX - MapMinX);
	xscale = (double)ScrMaxY / (double)(MapMaxY - MapMinY);
	if (xscale < yscale)
	{
		Scale = xscale * 0.8;
	}
	else
	{
		Scale = yscale * 0.8;
	}
}


/*
   center the map around the object and zoom in if necessary
   */

void GoToObject( BCINT objtype, BCINT objnum) /* SWAP! */
{
	BCINT xpos, ypos;
	BCINT xpos2, ypos2;
	BCINT n;
	BCINT sd1, sd2;
	float oldscale;

	// [RH] Switch to the editing mode appropriate for this object
	EditMode = objtype;

	GetObjectCoords( objtype, objnum, &xpos, &ypos);
	CenterMapAroundCoords( xpos, ypos);
	oldscale = Scale;

	/* zoom in until the object can be selected */
	while (Scale < 4.0 && GetCurObject( objtype, MAPX( PointerX ), MAPY( PointerY ), 4) != objnum)
	{
#if 0
		if (Scale < 1.0)
			Scale = 1.0 / ((1.0 / Scale) - 1.0);
		else
			Scale = Scale * 2.0;
#else
		Scale *= 1.0625;
#endif
	}
	/* Special case for Sectors: if several Sectors are one inside another, then	 */
	/* zooming in on the center won't help.  So I choose a LineDef that borders the  */
	/* the Sector and move a few pixels towards the inside of the Sector.			*/
	if (objtype == OBJ_SECTORS && GetCurObject( OBJ_SECTORS, OrigX, OrigY, 0) != objnum)
	{
		/* restore the Scale */
		Scale = oldscale;
		for (n = 0; n < NumLineDefs; n++)
		{
			sd1 = LineDefs[ n].sidedef1;
			sd2 = LineDefs[ n].sidedef2;
			if (sd1 >= 0 && SideDefs[ sd1].sector == objnum)
				break;
			if (sd2 >= 0 && SideDefs[ sd2].sector == objnum)
				break;
		}
		if (n < NumLineDefs)
		{
			GetObjectCoords( OBJ_LINEDEFS, n, &xpos2, &ypos2);
			n = ComputeDist( abs( xpos - xpos2), abs( ypos - ypos2)) / 7;
			if (n <= 1)
				n = 2;
			xpos = xpos2 + (xpos - xpos2) / n;
			ypos = ypos2 + (ypos - ypos2) / n;
			CenterMapAroundCoords( xpos, ypos);
			/* zoom in until the sector can be selected */
			while (Scale > 4.0 && GetCurObject( OBJ_SECTORS, OrigX, OrigY, 0) != objnum)
			{
#if 0
				if (Scale < 1.0)
					Scale = 1.0 / ((1.0 / Scale) - 1.0);
				else
					Scale = Scale / 2.0;
#else
				Scale *= 1.0625;
#endif
			}
		}
	}
	if (UseMouse)
		SetMouseCoords( PointerX, PointerY);
}
/* end of file */
