/*
   DETH - Doom Editor for Total Headcases, by Antony J. Burden
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.
   
   MAINLOOP.C - Editor routines.
   */

/* the includes */
#include <stdlib.h>
#include <time.h>
#ifndef __WATCOMC__
#include <pc.h>
#endif
#include <math.h>
#include <limits.h>
#include "deu.h"
#include "levels.h"
#include "keys.h"
#include <grx20.h>
#include <io.h>

void SortSelection(SelPtr *);
void DrawMenuLetters(BCINT);
void DoHelp(BCINT, BCINT);
void ShowHelpHere(BCINT);
void GetCount(SelPtr, BCINT);

void DrawArcSpots (int *circlex, int *circley, int circlepoints);

extern Bool InfoShown;		/* should we display the info bar? */
extern Bool BrowseMap;
Bool BMap;

/*
   the editor main loop
   */

#define REDRAW_NONE			0
#define REDRAW_QUICK		1
#define REDRAW_ALL			2
#define REDRAW_INFO			3
#define REDRAW_QUICKHOLD	4

#define CIRCLEPOINTS_MAX	128

// [RH] Make this global
BCINT EditMode;

void EditorLoop(Bool browsed)
{
#ifdef __WATCOMC__
	clock_t lastscroll = 0;
#else
	struct timeval lastscroll = { 0 };
#endif
	Bool showZeroMessage = FALSE;
    SelPtr Selected = NULL;
    SelPtr n;
    BCINT CurObject = -1;
    BCINT OldObject = -1;
	BCINT RedrawMap = REDRAW_ALL;
    BCINT LastRedraw = REDRAW_NONE;
    BCINT buttons; 
    BCINT oldbuttons = 0;
    BCINT SelBoxX = 0;
    BCINT SelBoxY = 0;
    BCINT OldPointerX = 0;
    BCINT OldPointerY = 0;
    BCINT val = 0;
    BCINT count = 0;
    BCINT x0, y0; 
    BCINT MaxX = ScrMaxX;
    BCINT MaxY = ScrMaxY;
    BCINT sd1x, sd2x, sd1y, sd2y;
    Bool  RedrawObj = FALSE;
    Bool  InfoDisplayed = FALSE;
    Bool  LettersDrawn = FALSE;		/*FULL*/
    Bool  StretchSelBox = FALSE;
    Bool  KeepMenu = FALSE;
    Bool  DrawMode = FALSE;
    int   key, altkey;
    char  keychar;

	// [RH] Arc-drawing stuff
	int   circleline = -1;
	int   lastcircleline = -1;
	int   circlepoints = 7;
	Bool  circlebad = TRUE;
	int	  circlex[CIRCLEPOINTS_MAX+1], circley[CIRCLEPOINTS_MAX+1];

	static const BCINT mbKeys[] =
	{
		0x2100, // F
		0x1200, // E
		0x1f00, // S
		0x3200, // M
		0x1700, // I
		0x1800, // O
		0x2E00, // C
		0x2300, // H
	};

	EditMode = EdMode;
    DraggingObject = FALSE; 
    MadeChanges = FALSE;
    MadeMapChanges = FALSE;
    if (InitialScale > 40)
		InitialScale = 40;
    Scale = (float) (1.0 / InitialScale);
	CenterMapAroundCoords((MapMinX + MapMaxX) / 2, (MapMinY + MapMaxY) / 2);
    if (UseMouse)
	{
		ResetMouseLimits();
		SetMouseCoords( PointerX, PointerY);
		ShowMousePointer();
		oldbuttons = 0;
		// [RH] Center the mouse pointer so the map doesn't go scrolling away
		SetMouseCoords (ScrMaxX/2, ScrMaxY/2);
    }
    else
	{
		FakeCursor = TRUE;
	}
	if (CreatingNewLevel)
	{
		InsertObject( OBJ_SECTORS, -1, 0, 0);
		CreatingNewLevel = FALSE;
	}
	else if (!browsed && BehaviorSize == 0 && NonHexenWarning)
	{
		showZeroMessage = TRUE;
	}

    // [RH] The map has just been loaded, so don't let the user undo to
	// a previous map.
	last_backup = 0;
	strcpy (BackupText, "Nothing To Undo");

    for (;;)
	{
		key = 0;
		altkey = 0;
		
		/* get mouse position and button status */
		if (UseMouse)
		{
			if (FakeCursor)
			{
				HideMousePointer();
				DrawPointer();
				ShowMousePointer();
			}
			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (FakeCursor)
			{
				HideMousePointer();
				DrawPointer();
				ShowMousePointer();
			}
			MaxX = ScrMaxX;
			sd1x = ScrMaxX - 140;
			sd1y = ScrMaxY - 266;
			sd2x = ScrMaxX - 72;
			sd2y = ScrMaxY - 266;
			if (VideoMode < 3)
			{
				sd1y -= 91;
				sd2y -= 91;
			}
			if (InfoShown)
			{
				MaxY = ScrMaxY;
				sd1y -= 13;
				sd2y -= 13;
			}
			else
			{
				MaxY = ScrMaxY + 13;
			}
			if (buttons == 1 && PointerY < 16)
			{
				/* kludge for the menu bar */
				int slot = -1;

				altkey = MOD_ALT;
				if (PointerX < 12)			DEU_Beep();
				else if (PointerX < 42)		slot = 0; /* 'F' */
				else if (PointerX < 80)		slot = 1; /* 'E' */
				else if (PointerX < 135)	slot = 2; /* 'S' */
				else if (PointerX < 176)	slot = 3; /* 'M' */
				else if (PointerX < 247)	slot = 4; /* 'I' */
				else if (PointerX < 344)	slot = 5; /* 'O' */
				else if (PointerX < 440)	slot = 6; /* 'C' */
				else if (PointerX < ScrMaxX - 60)	DEU_Beep();
				else						slot = 7; /* 'H' */
				if (slot >= 0)
				{
					key = mbKeys[slot];
				}
			}
			if (buttons != oldbuttons && CurObject >= 0)
			{
				switch (buttons)
				{
				case 1:
					if (SwapButtons)
					{
						key = ENTER;
					}
					else
					{
						if (DrawMode && (EditMode == OBJ_VERTEXES || EditMode == OBJ_THINGS))
							key = INSERT;
						else if (circleline >= 0)
							key = INSERT;
						else
							key = 'M';
					}
					break;
				case 2:
					if (!DraggingObject)
					{
						if (EditMode == OBJ_THINGS)
						{
							Backup (BTHINGS);
							strcpy (BackupText, "Move Things");
						}
						else if (EditMode == OBJ_VERTEXES)
						{
							Backup (BVERTEXES | BLINEDEFS | BLINEDEFS);
							strcpy (BackupText, "Move Vertexes");
						}
						else if (EditMode == OBJ_LINEDEFS)
						{
							Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS);
							strcpy (BackupText, "Move LineDefs");
						}
						else if (EditMode == OBJ_SECTORS)
						{
							Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS);
							strcpy (BackupText, "Move Sectors");
						}
						key = 'D';	  /* Press right button = Drag */
					}
					break;
				case 3:
				case 4:
					if (SwapButtons)
					{
						if (DrawMode && (EditMode == OBJ_VERTEXES || EditMode == OBJ_THINGS))
							key = INSERT;
						else if (circleline >= 0)
							key = INSERT;
						else
							key = 'M';
					}
					else
					{
						key = ENTER;   /* Press middle button = Edit ('Enter') */
					}
					break;
				default:
					if (StretchSelBox) /* Release left button = End Selection Box */
					{
						key = 'M';
					}
					if (DraggingObject)
					{	/* Release right button = End Drag */
						DisplayObjectInfo( EditMode, CurObject);
						DrawInfoBar(EditMode, GridScale);
						key = 'D'; 
					}
					break;
				}
				altkey = bioskey (2);
			}
			oldbuttons = buttons;
		}
		if (InfoDisplayed && RedrawMap != REDRAW_ALL && !StretchSelBox && !DraggingObject && CurObject >= 0)
		{
			x0 = ScrMaxX - 140;
			y0 = ScrMaxY - 210;
			if (VideoMode < 3)
				y0 -= 91;
			if (InfoShown) 
				y0 -= 13; 
			if (EditMode == OBJ_THINGS && MouseInArea(2, MaxY - 132, 360, 116))
				RedrawMap = OnScreenEditingThings(CurObject, Selected);
			else if (DispSprite && EditMode == OBJ_THINGS && MouseInArea(MaxX - 120, MaxY - 132, 116, 120))
			{
				ShowHelpHere(ORANGE);
				RedrawMap = OnScreenEditingThings(CurObject, Selected);
				ShowHelpHere(WHITE);
			}
			else if (EditMode == OBJ_LINEDEFS && MouseInArea(2, MaxY - 142, 258, 120))
			{
				RedrawMap = OnScreenEditingLineDefs(CurObject, Selected);
			}
			else if (EditMode == OBJ_LINEDEFS && DispSprite && LineDefs[CurObject].sidedef1 > -1 && MouseInArea(x0, y0, 136, 210))
			{
				ShowHelpHere(ORANGE);
				RedrawMap = OnScreenEditingSideDefTextures(CurObject, Selected);
				ShowHelpHere(WHITE);
			}
			else if (EditMode == OBJ_SECTORS && MouseInArea(2, MaxY - 92, 280, 76))
			{
				RedrawMap = OnScreenEditingSectors(CurObject, Selected);
			}
			else if (EditMode == OBJ_SECTORS && DispSprite && MouseInArea(MaxX - 72, MaxY - 148, 72, 140))
			{
				ShowHelpHere(ORANGE);
				RedrawMap = OnScreenEditingSectorFlats(CurObject, Selected);
				ShowHelpHere(WHITE);
			}
			else while (EditMode == OBJ_LINEDEFS && MouseInArea(260, MaxY - 96, 380, 80))
			{
				RedrawMap = REDRAW_NONE;
				GetMouseCoords( &PointerX, &PointerY, &buttons);
				if (LineDefs[CurObject].sidedef1 > -1 && MouseInArea(260, MaxY - 96, 192, 80))
				{
					RedrawMap = OnScreenEditingSideDef1(CurObject, Selected);
					if (RedrawMap > REDRAW_NONE)
						break;
				}
				else if (LineDefs[CurObject].sidedef2 > -1 && MouseInArea(450, MaxY - 96, 190, 80))
				{
					RedrawMap = OnScreenEditingSideDef2(CurObject, Selected);
					if (RedrawMap > REDRAW_NONE)
						break;
				}
				else if (LineDefs[CurObject].sidedef1 == -1 && MouseInArea(304, MaxY - 43, 100, 16) && buttons == 1)
				{ 
					strcpy (BackupText, "Add 1st SideDef");
					Backup (BSIDEDEFS | BLINEDEFS);
					for (n = Selected; n; n = n->next)
					{
						if (LineDefs[ n->objnum].sidedef1 == -1)
						{
							InsertObject( OBJ_SIDEDEFS, -1, 0, 0);
							LineDefs[ n->objnum].sidedef1 = NumSideDefs - 1;
						}
					}
					if (LineDefs[CurObject].sidedef1 == -1)
					{
						InsertObject( OBJ_SIDEDEFS, -1, 0, 0);
						LineDefs[CurObject].sidedef1 = NumSideDefs - 1;
					}
					HideMousePointer();
					DrawScreenButtonIn(304, MaxY - 43, 100, 16, "Add 1st SideDef", TRUE);
					ShowMousePointer();
					RedrawMap = REDRAW_ALL;
					break;
				}
				else if (LineDefs[CurObject].sidedef2 == -1 && MouseInArea(495, MaxY - 43, 100, 16) && buttons == 1)
				{ 
					DrawScreenButtonIn(495, MaxY - 43, 100, 16, "Add 2nd SideDef", TRUE);
					strcpy (BackupText, "Add 2nd SideDef");
					Backup (BSIDEDEFS | BLINEDEFS);
					for (n = Selected; n; n = n->next)
					{
						if (LineDefs[ n->objnum].sidedef2 == -1)
						{
							val = LineDefs[ n->objnum].sidedef1;
							InsertObject( OBJ_SIDEDEFS, -1, 0, 0);
							strncpy( SideDefs[ NumSideDefs - 1].tex3, "-", 8);
							strncpy( SideDefs[ val].tex3, "-", 8);
							LineDefs[ n->objnum].sidedef2 = NumSideDefs - 1;
							LineDefs[ n->objnum].flags = 4;
						}
					}
					if (LineDefs[CurObject].sidedef2 == -1)
					{
						val = LineDefs[CurObject].sidedef1;
						InsertObject( OBJ_SIDEDEFS, -1, 0, 0);
						strncpy( SideDefs[ NumSideDefs - 1].tex3, "-", 8);
						strncpy( SideDefs[ val].tex3, "-", 8);
						LineDefs[CurObject].sidedef2 = NumSideDefs - 1;
						LineDefs[CurObject].flags = 4;
					}
					HideMousePointer();
					DrawScreenButtonIn(495, MaxY - 43, 100, 16, "Add 2nd SideDef", TRUE);
					ShowMousePointer();
					RedrawMap = REDRAW_ALL;
					break;
				}
			}
		}

		/* drag object(s) */
		if (DraggingObject)
		{
			BCINT forgetit = FALSE;
			
			if (IsSelected( Selected, CurObject) == FALSE)
			{
				ForgetSelection( &Selected);
			}
			else if (Selected->objnum != CurObject)
			{
				/* current object must be first in the list */
				UnSelectObject( &Selected, CurObject);
				SelectObject1( &Selected, CurObject);
			}
			if (Selected == NULL && CurObject >= 0)
			{
				SelectObject1( &Selected, CurObject);
				forgetit = TRUE;
			} 
			if (Selected)
			{
				GetCount(Selected, forgetit ? -1 : 0);
				if (MoveObjectsToCoords( EditMode, Selected, MAPX( PointerX), MAPY( PointerY), GridScale))
					RedrawMap = REDRAW_ALL;
				if (forgetit)
					ForgetSelection( &Selected);
			}
			else
			{
				DEU_Beep();
				DraggingObject = FALSE;
			}
		}
		else if (StretchSelBox)
		{
			BCINT x = MAPX( PointerX);
			BCINT y = MAPY( PointerY);
			
			/* draw selection box */
			SetColor( CYAN);
			SetLineDotted (TRUE);
			SetWriteMode( XOR_PUT);
			DrawMapHollowBox (SelBoxX, SelBoxY, x, y);
			delay(25);
			DrawMapHollowBox (SelBoxX, SelBoxY, x, y);
			SetWriteMode( COPY_PUT);
			SetLineDotted (FALSE);
			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (buttons == 0)
				key = 'M';
		}
		else if (!RedrawObj && circleline < 0)
		{
			if ((bioskey (2) & MOD_SHIFT) == 0x00)  /* no shift keys */
			{
				/* check if there is something near the pointer */
				OldObject = CurObject;/*
				if (Scale >= 4.0)
					Sensitivity = 10;
				else if (Scale >= 2.0)
					Sensitivity = 8;
				else if (Scale >= 1.0)
					Sensitivity = 6;
				else if (Scale >= 0.5)
					Sensitivity = 5;
				else if (Scale >= 0.4)
					Sensitivity = 4;
				else if (Scale >= 0.3)
					Sensitivity = 3;
				else if (Scale >= 0.2)
					Sensitivity = 2;
				else
					Sensitivity = 1;*/
				// [RH] Why the above scale?
				Sensitivity = 10.0 / Scale;
				if (Sensitivity < 1.0)
					Sensitivity = 1.0;
					CurObject = GetCurObject( EditMode, MAPX( PointerX ), MAPY( PointerY ), Sensitivity);
				if (CurObject < 0)
					CurObject = OldObject;
			}
		}



		/* draw the map */

		// [RH] Clamp the origin so the user can't scroll way off the map
		if (OrigX < -32768)
			OrigX = -32768;
		else if (OrigX > 32767)
			OrigX = 32767;
		if (OrigY < -32768)
			OrigY = -32768;
		else if (OrigY > 32767)
			OrigY = 32767;

		// [RH] This is wrapped inside a do loop so that if the "no behavior" message
		// is presented, the map will redraw when the user dismisses it.
		do
		{
			if(RedrawMap == REDRAW_NONE && (LastRedraw == REDRAW_QUICK || LastRedraw == REDRAW_QUICKHOLD))
				RedrawMap = REDRAW_ALL;

			if (RedrawMap == REDRAW_ALL)
			{
				GetCount(Selected, 0);
				HideMousePointer();
				DrawMap(EditMode, GridScale, GridShown);
				ShowMousePointer();
				HighlightSelection(EditMode, Selected);
				lastcircleline = -1;
			}
			else if(RedrawMap == REDRAW_INFO && InfoDisplayed)
			{
				DisplayObjectInfo(EditMode, CurObject);
				DrawInfoBar(EditMode, GridScale);
			}
			else if(RedrawMap == REDRAW_QUICK)
			{
				GetCount(Selected, 0);
				HideMousePointer();
				if (!DetailScroll)
				{
					DrawMapQuick();
				}
				else
				{
					DrawMap(EditMode, GridScale, FALSE);
				}
				ShowMousePointer();
				InfoDisplayed = FALSE;
				lastcircleline = -1;
			}

			if (DrawMode && (RedrawMap == REDRAW_ALL || RedrawMap == REDRAW_QUICK))
			{
				SetColor(YELLOW);
				DrawScreenText(ScrMaxX - 220, 5, "DM");
			}


			/* highlight the current object and display the information box */
			if ((CurObject != OldObject || RedrawObj || RedrawMap == REDRAW_ALL) && PointerX < ScrMaxX && PointerX > 0 && PointerY < ScrMaxY && PointerY > 0)
			{
				RedrawObj = FALSE;
				if (!RedrawMap && OldObject >= 0)
				{
					HighlightObject( EditMode, OldObject, YELLOW); 
				}
				if (CurObject != OldObject)
				{
					PlaySound( 50, 10);
					OldObject = CurObject;
					if (EditMode == OBJ_SECTORS && Sectors[CurObject].tag > 0)
					{
						LastSectorTag = Sectors[CurObject].tag;
					}
					if (EditMode == OBJ_LINEDEFS)
					{
						char *text = GetLineDefArgs (LineDefs[CurObject].special);
						int i;

						for (i = 0; i < 5; ++i)
						{
							if ((text[i] == '$' || text[i] == '%') && LineDefs[CurObject].arg[i])
							{
								LastLineDefTag = LineDefs[CurObject].arg[i];
								break;
							}
						}
					}
				}
				if (CurObject >= 0)
				{
					HighlightObject( EditMode, CurObject, YELLOW);
				}
				if (bioskey (1)) /* speedup */
				{
					RedrawObj = TRUE;
				}
				else if (!DraggingObject)
				{
					DisplayObjectInfo(EditMode, CurObject);
					InfoDisplayed = TRUE;
				}
			}
			if (RedrawMap && FakeCursor)
			{
				HideMousePointer();
				DrawPointer();
				ShowMousePointer();
			}

			/* display the current pointer coordinates */
			if (RedrawMap == REDRAW_ALL || PointerX != OldPointerX || PointerY != OldPointerY)
			{
				SetColor(DARKGRAY);
				DrawScreenBox( ScrMaxX - 185, 5, ScrMaxX - 70, 12);
				SetColor( BLUE);
				DrawScreenText( ScrMaxX - 185, 5, "(%d, %d)", (int)MAPX( PointerX), (int)MAPY( PointerY));
				if (Debug)
				{
					SetColor( BLACK);
					DrawScreenBox( ScrMaxX - 185, 25, ScrMaxX - 70, 32);
					SetColor( ORANGE);
					DrawScreenText( ScrMaxX - 185, 25, "(%d, %d ,%d, %2f)", PointerX, ScrMaxY - PointerY, Sensitivity, Scale);
				}
				OldPointerX = PointerX;
				OldPointerY = PointerY;
			}

			if (showZeroMessage)
			{
				Notify (-1, -1, "This map is not in the Hexen format, so all its specials have been zeroed.",
								"Use ZWADCONV to convert the map if you want to keep them.");
			}
		} while (showZeroMessage && !(showZeroMessage = FALSE));

		/* the map is up to date */
		LastRedraw = RedrawMap;
		RedrawMap = REDRAW_NONE;
		if (bioskey (2) == (MOD_NUMLOCK|MOD_ALT) && !LettersDrawn)
		{
			DrawMenuLetters(BLACK);
			LettersDrawn = TRUE;
		}
		if (bioskey (2) != (MOD_NUMLOCK|MOD_ALT) && LettersDrawn)
		{
			DrawMenuLetters(WHITE);
			LettersDrawn = FALSE;
		}
		/* [RH] add arc dragging here */
		if (circleline >= 0 && EditMode == OBJ_LINEDEFS)
		{
			int x1 = Vertexes[LineDefs[circleline].start].x;
			int y1 = Vertexes[LineDefs[circleline].start].y;
			int x2 = Vertexes[LineDefs[circleline].end].x;
			int y2 = Vertexes[LineDefs[circleline].end].y;
			int x3 = MAPX(PointerX);
			int y3 = MAPY(PointerY);

			double a = x2 - x1;
			double b = y2 - y1;
			double c = x3 - x1;
			double d = y3 - y1;
			double e = a * (x1 + x2) + b * (y1 + y2);
			double f = c * (x1 + x3) + d * (y1 + y3);
			double g = a * (y3 - y2) - b * (x3 - x2);

			if (g != 0.0)
			{
				int i;
				double rad;
				double cx, cy;
				double ang1, ang2;

				g = 0.5 / g;
				cx = (d*e - b*f) * g;
				cy = (a*f - c*e) * g;
				rad = sqrt ((cx-x1)*(cx-x1) + (cy-y1)*(cy-y1));
				ang1 = atan2 (y1 - cy, x1 - cx);
				ang2 = atan2 (y2 - cy, x2 - cx);

				if (ang1 < 0)
				{
					ang1 = 2*M_PI + ang1;
				}
				if (ang2 < 0)
				{
					ang2 = 2*M_PI + ang2;
				}

				if (c*b < a*d)
				{
					if (ang1 < ang2)
					{
						ang1 += 2*M_PI;
					}
				}
				else if (ang2 < ang1)
				{
					ang2 += 2*M_PI;
				}

				if (lastcircleline < 0 || (int)cx != circlex[0] || (int)cy != circley[0] || circlebad)
				{
					SetColor( LIGHTGREEN);
					SetWriteMode (XOR_PUT);
					if (lastcircleline >= 0 && !circlebad)
					{
						DrawArcSpots (circlex, circley, circlepoints);
					}
					circlex[0] = (int)cx;
					circley[0] = (int)cy;

					for (i = 0; i < circlepoints; ++i)
					{
						double ang = ang1 + (ang2 - ang1) * (i+1) / (circlepoints+1);
						int px = (int)(cx + cos(ang) * rad);
						int py = (int)(cy + sin(ang) * rad);

						if (px < -32768)
							px = -32768;
						else if (px > 32767)
							px = 32767;
						if (py < -32768)
							py = -32768;
						else if (py > 32767)
							py = 32767;

						circlex[i+1] = px;
						circley[i+1] = py;
					}

					DrawArcSpots (circlex, circley, circlepoints);
					lastcircleline = circleline;
					SetWriteMode (COPY_PUT);
				}
				circlebad = FALSE;
			}
			else
			{
				if (lastcircleline >= 0)
				{
					SetColor (LIGHTGREEN);
					SetWriteMode (XOR_PUT);
					DrawArcSpots (circlex, circley, circlepoints);
					SetWriteMode (COPY_PUT);
					lastcircleline = -1;
				}
				circlebad = TRUE;
			}
		}
		else if (lastcircleline >= 0)
		{
			SetColor (LIGHTGREEN);
			SetWriteMode (XOR_PUT);
			DrawArcSpots (circlex, circley, circlepoints);
			SetWriteMode (COPY_PUT);
			lastcircleline = -1;
		}
		else if (circleline >= 0)
		{
			circleline = -1;
		}

		
		/* get user input */
		if (bioskey (1) || key)
		{
			if (!key)
			{
				key = bioskey (0);
				altkey = bioskey (2);
			}
			
			if (Debug)
			{
				SetColor(BLACK);
				DrawScreenBox( ScrMaxX - 300 , 25, ScrMaxX - 250, 35);  
				SetColor(ORANGE);
				DrawScreenText( ScrMaxX - 300 , 25, "%d, %d", key, altkey);  
			}


			/* user wants to access the drop-down menus */
			if (altkey & MOD_ALT)   /* if alt is pressed... */
			{
				KeepMenu = TRUE;
				DrawMenuLetters(WHITE);
				while (KeepMenu)
				{
					x0 = -1;
					y0 = NestMenus ? 17 : -1;
					if ((key & 0xFF00) == 0x2100)	   /* Scan code for F */
					{
						if (NestMenus)
						{
							x0 = 0;
						}
						key = PullDownMenu( x0, y0,
										   "Save           F2", 0x3C00,    (BCINT) 'S', 1,
										   "Save As...     F3", 0x3D00,    (BCINT) 'A', 6,
										   "-", 				0x3D00,    (BCINT) 'A', 1,
										   "Preferences    F5", F5    ,    (BCINT) 'P', 1,
										   "Save ini File    ",	CTRL_6,    (BCINT) 'I', 6,
										   "-", 				0x3D00,    (BCINT) 'A', 1,
										   "Drop to DOS      ",	ESCAPE,    (BCINT) 'D', 1,
										   "-", 				0x3D00,    (BCINT) 'A', 1,
										   "Quit             ", CTRL_W,    (BCINT) 'T', 4,
										   "Quit & Save     Q", CTRL_Q,    (BCINT) 'Q', 1,
										   NULL); 
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x1200; /* 'E' */
							}
							else if (key == -2)
							{
								key = 0x2300; /* 'H' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						KeepMenu = FALSE;
					}
					else if ((key & 0xFF00) == 0x1200)  /* Scan code for E */
					{
						if (NestMenus)
						{
							x0 = 43;
						}
						key = PullDownMenu( x0, y0,
											"Copy Object(s)       C", (BCINT) 'O', (BCINT) 'C', 1,
											"Add Object         Ins", INSERT,      (BCINT) 'A', 1,

											"Delete Object(s)   Del", DELETE,      (BCINT) 'D', 1,
											"-", 					  0x3D00,      (BCINT) 'A', 1,
											"Edit Scripts",			  0x7eef,      (BCINT) 'E', 1,
											"-", 					  0x3D00,      (BCINT) 'A', 1,
											"Preferences         F5", F5    ,      (BCINT) 'P', 1,
											"Display Options     F7", F7    ,      (BCINT) 'O', 9,
										   NULL);
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x1F00; /* 'S' */
							}
							else if (key == -2)
							{
								key = 0x2100; /* 'F' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						KeepMenu = FALSE;
					}
					else if ((key & 0xFF00) == 0x1F00)  /* Scan code for S */
					{
						if (NestMenus)
						{
							x0 = 81;
						}
						if (EditMode == OBJ_THINGS)
							key = PullDownMenu( x0, y0,
										   "Find/Mark/Filter...  F", (BCINT) 'F', (BCINT) 'F', 1,
										   "Find Again           A", (BCINT) 'A', (BCINT) 'A', 6,
											"-", 					 0x3D00,      (BCINT) 'A', 1,
										   "Next Object          N", (BCINT) 'N', (BCINT) 'N', 1,
										   "Prev Object          P", (BCINT) 'P', (BCINT) 'P', 1,
										   "Jump To Object...    J", (BCINT) 'J', (BCINT) 'J', 1,
										   NULL);
						else
							key = PullDownMenu( x0, y0,
										   "Next Object          N", (BCINT) 'N', (BCINT) 'N', 1,
										   "Prev Object          P", (BCINT) 'P', (BCINT) 'P', 1,
										   "Jump To Object...    J", (BCINT) 'J', (BCINT) 'J', 1,
										   NULL);
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x3200; /* 'M' */
							}
							else if (key == -2)
							{
								key = 0x1200; /* 'E' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						KeepMenu = FALSE;
					}
					else if ((key & 0xFF00) == 0x3200)  /* Scan code for M */
					{
						if (NestMenus)
						{
							x0 = 136;
						}
						key = PullDownMenu( x0, y0,
										   ((EditMode == OBJ_THINGS) ?
											"* Things              T" :
											"  Things              T"), (BCINT) 'T', (BCINT) 'T', 3,
										   ((EditMode == OBJ_LINEDEFS) ?
											"* LineDefs+SideDefs   L" :
											"  LineDefs+SideDefs   L"), (BCINT) 'L', (BCINT) 'L', 3,
										   ((EditMode == OBJ_VERTEXES) ?
											"* Vertexes            V" :
											"  Vertexes            V"), (BCINT) 'V', (BCINT) 'V', 3,
										   ((EditMode == OBJ_SECTORS) ?
											"* Sectors             S" :
											"  Sectors             S"), (BCINT) 'S', (BCINT) 'S', 3,
											"-", 						0x3D00,      (BCINT) 'S', 1,
											"Next Mode           Tab",  0x0009,      (BCINT) 'N', 1,
											"Last Mode     Shift+Tab",  0x0F00,      (BCINT) 'L', 1,
										   NULL);
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x1700; /* 'I' */
							}
							else if (key == -2)
							{
								key = 0x1F00; /* 'S' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						KeepMenu = FALSE;
					}
					else if ((key & 0xFF00) == 0x1700) /* Scan code for I */
					{
						if (NestMenus)
						{
							x0 = 177;
						}
						key = 0;
						/* code duplicated from 'F8' - I hate to do that */
						if (Selected) /* JFF mis-blocked code */
						{
							key = MiscOperations( x0, y0, EditMode, &Selected, CurObject);
						}/* JFF mis-blocked code */
						else
						{						// jff Don't select current
							if (CurObject >= 0) // when replacing textures
							{
								SelectObject1( &Selected, CurObject); // allow empty to mean all
							}
							key = MiscOperations( x0, y0, EditMode, &Selected, CurObject);
							if (CurObject >= 0)
							{
								UnSelectObject( &Selected, CurObject);
							}
						}
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x1800; /* 'O' */
							}
							else if (key == -2)
							{
								key = 0x3200; /* 'M' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						CurObject = -1;
						DraggingObject = FALSE;
						StretchSelBox = FALSE;
						GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
						KeepMenu = FALSE;
					}
					else if ((key & 0xFF00) == 0x1800) /* Scan code for O */
					{
						BCINT savednum, i;
						if (NestMenus)
						{
							x0 = 248;
						}
						
						key = 0;
						/* don't want to create the object behind the menu bar... */
						if (PointerY < 20)
						{
							PointerX = ScrCenterX;
							PointerY = ScrCenterY;
						}
						/* code duplicated from 'F9' - I hate to do that */
						savednum = NumLineDefs;
						key = InsertStandardObject( x0, y0, MAPX( PointerX), MAPY( PointerY));
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x2E00; /* 'C' */
							}
							else if (key == -2)
							{
								key = 0x1700; /* 'I' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						if (NumLineDefs > savednum)
						{
							ForgetSelection( &Selected);
							EditMode = OBJ_LINEDEFS;
							for (i = savednum; i < NumLineDefs; i++)
							{
								SelectObject1( &Selected, i);
							}
							CurObject = NumLineDefs - 1;
							OldObject = -1;
							DraggingObject = FALSE;
							StretchSelBox = FALSE;
							GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
						}
						KeepMenu = FALSE;
					}
					else if ((key & 0xFF00) == 0x2E00) /* Scan code for C */
					{
						if (NestMenus)
						{
							x0 = 344;
						}
						key = CheckLevel( x0, y0);
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x2300; /* 'H' */
							}
							else if (key == -2)
							{
								key = 0x1800; /* 'O' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						KeepMenu = FALSE;
					}
					else if ((key & 0xFF00) == 0x2300)  /* Scan code for H */
					{
						if (NestMenus) {
							x0 = ScrMaxX - 100;
						}
						key = PullDownMenu( x0, y0,
							 "General Keys    F1", F1,		  (BCINT) 'G', 1,
							 "Special Keys    F6", F6,		  (BCINT) 'S', 1,
							"-", 				   0x3D00,    (BCINT) 'S', 1,
							 "About...",		   SHIFT_F10, (BCINT) 'A', 1,
							  NULL);
						if (NestMenus && key < 0)
						{
							if (key == -1)
							{
								key = 0x2100; /* 'F' */
							}
							else if (key == -2)
							{
								key = 0x2E00; /* 'C' */
							}
							else
							{
								key = mbKeys[-10-key];
							}
							KeepMenu = TRUE;
							DrawMap(EditMode, GridScale, GridShown);
							DisplayObjectInfo(EditMode, CurObject);
							continue;
						}
						KeepMenu = FALSE;
					}
					else
					{
						DEU_Beep();
						key = 0;
					}
					KeepMenu = FALSE;
					RedrawMap = REDRAW_ALL;
				}
			}
			
			
			/* simplify the checks later on */
			if (isprint(key & 0x00ff)) 
				keychar = toupper(key & 0x00ff);
			else 
				keychar = '\0';

			
			/* erase the (keyboard) pointer */
			if (FakeCursor)
			{
				HideMousePointer();
				DrawPointer();
				ShowMousePointer();
			}
			
			if(keychar == 'F' && EditMode == OBJ_THINGS)
			{
				ShowSprite = TRUE;
				FindThing(0, &Selected);
				ShowSprite = FALSE;
				RedrawMap = REDRAW_ALL;
			}

			else if(keychar == 'A' && EditMode == OBJ_THINGS)
			{
				ShowSprite = TRUE;
				FindThing(1, NULL);
				ShowSprite = FALSE;
				RedrawMap = REDRAW_ALL;
			}

			/* [RH] make an arc out of a line ala Build */
			else if(keychar == 'A' && EditMode == OBJ_LINEDEFS)
			{
				if (circleline >= 0)
				{
					circleline = -1;
				}
				else if (CurObject >= 0)
				{
					circleline = CurObject;
				}
			}
			
			/* user wants to exit  */
			else if (IS_CTRL_KEY(key, 'W') || (QisQuit && keychar == 'Q'))
			{
				if (!MadeChanges || Confirm(-1, -1, "You Have Unsaved Changes.  Do You Really Want To Quit?", NULL))
				{
					SaveConfigFileOptions();
					break;
				}
				RedrawMap = REDRAW_ALL;
			} 
			else if ((key & 0x00FF) == 0x001B) /* 'Esc' */
			{
				if (DraggingObject)
				{
					DraggingObject = FALSE;
				}
				else if (StretchSelBox)
				{
					StretchSelBox = FALSE;
					GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
				}
				else
				{
					ForgetSelection( &Selected);
					if (!MadeChanges || Confirm(-1, -1, "You Have Unsaved Changes.  Do You Really Want To Quit?", NULL))
					{
						SaveConfigFileOptions();
						break;
					}
					RedrawMap = REDRAW_ALL;
				}
			}

			/* user is lost */
			else if (key == F1)
			{
				DoHelp(EditMode, GridScale);
				RedrawMap = REDRAW_ALL;
			}
			
			/* user wants special keys list */
			else if (key == F6)
			{ 
				DisplayHotKeys(EditMode);
				RedrawMap = REDRAW_ALL;
			}

			/* user wants the about box */
			else if (key == SHIFT_F10)
			{ 
				DisplayAbout();
				RedrawMap = REDRAW_ALL;
			}

			/* user wants to edit the map's scripts */
			else if (key == (BCINT)0x7eef)
			{
				int really = 1;
				
				if (BehaviorSize && !ScriptsSize)
				{
					int x0 = (ScrMaxX - 22 - 8*50) / 2;
					int y0 = (ScrMaxY - 63) / 2;
					int blah;
					really = 0;
					DrawScreenBox3D (x0, y0, x0 + 22 + 8*50, y0 + 73);
					SetColor (WHITE);
					DrawScreenText (x0+10, y0+8, "This map did not have a SCRIPTS lump.");
					DrawScreenText (x0+10, y0+18, "Any existing scripts will be lost if you continue.");
					DEU_Beep ();
					SetColor (YELLOW);
					DrawScreenText (x0+10, y0+38, "Press Y if you really want to do this");
					DrawScreenText (x0+10, y0+48, "or any other key to cancel.");
					blah = bioskey (0);
					if ((blah & 0xff) == 0x79)
						really = 1;
				}
				if (really)
				{
					char *cwd;

					cwd = getcwd (NULL, PATH_MAX);
					if (cwd)
					{
						if (0 == chdir (ACCDirectory))
						{
							static const char tempfile[] = "zethtemp";
							char buff[1024];
							FILE *file;
							int i;

							// Make sure if we do get an error log, it's for this script file
							remove ("acs.err");

							sprintf (buff, "%s.acs", tempfile);
							file = fopen (buff, "wb");
							if (ScriptsSize)
								fwrite (Scripts, 1, ScriptsSize, file);
							else
								fwrite ("#include \"zcommon.acs\"\n\n", 1, 24, file);
							fclose (file);
							_chmod (buff, 1, 0);
							TermGfx ();

edit:
							sprintf (buff, "%s %s.acs", TextEditor, tempfile);
							for (i = 0; buff[i]; i++)
								if (buff[i] == '/')
									buff[i] = '\\';

							printf ("%s\n", buff);
							system (buff);

							// Delay before compiling in case editor returns
							// immediately (such as Notepad) - Ty 2000/01/29
							// v4.04, Readded by Ty 2002/12/08 for v4.14t
							// [RH] Actually, any Windows editor will do this
							// because they got spawned as separate processes,
							// and there's no way for a DOS-based task to know
							// about their lifetime.
							if (PauseBeforeCompile)
							{
								printf ("\nPress Y when ready to compile or any other key\nto reinvoke the text editor.\n");
								fflush (stdout);
								i = getkey ();
								if (toupper(i) != 'Y')
									goto edit;
							}

							sprintf (buff, "%s %s.acs", ACCCommand, tempfile);
							for (i = 0; buff[i]; i++)
								if (buff[i] == '/')
									buff[i] = '\\';

							printf ("%s\n", buff);
							system (buff);

							file = fopen ("acs.err", "r");
							if (file)
							{
								char line[1024];
								int lines, totrows;

								// something went wrong
								ScreenClear ();
								ScreenSetCursor (0,0);
								totrows = ScreenRows () - 1;
								while (1)
								{
									for (lines = 0; lines < totrows && fgets (line, 1023, file); lines++)
										fputs (line, stdout);

									if (lines == totrows)
									{
										printf ("--- More ---\r");
										fflush (stdout);
										getkey ();
										printf ("            \r");
									}
									else
										break;
								}
								fclose (file);

								printf ("\nDo you want to edit the scripts again? ");
								fflush (stdout);
								i = getkey ();
								if (toupper(i) == 'Y')
									goto edit;
							}
							else
							{
								// the scripts were compiled successfully
								printf ("\nDo you want to use the new behavior? ");
								fflush (stdout);
								i = getkey ();
								while (toupper(i) == 'Y')
								{
									void *newscripts;
									void *newbehavior;
									int newscriptssize;
									int newbehaviorsize;

									sprintf (buff, "%s.acs", tempfile);
									FileRead (buff, &newscripts, &newscriptssize);
									if (newscripts)
									{
										sprintf (buff, "%s.o", tempfile);
										FileRead (buff, &newbehavior, &newbehaviorsize);
										if (newbehavior)
										{
											if (ScriptsSize)
												free (Scripts);
											if (BehaviorSize)
												free (Behavior);

											ScriptsSize = newscriptssize;
											Scripts = newscripts;
											BehaviorSize = newbehaviorsize;
											Behavior = newbehavior;
											MadeChanges = TRUE;
										}
										else
										{
											free (newscripts);
											newscripts = NULL;
										}
									}
									if (newscripts)
										break;

									printf ("\nSomething went wrong. Try again? ");
									fflush (stdout);
									i = getkey ();
								}
							}

							printf ("\n");
							sprintf (buff, "%s.acs", tempfile);
							remove (buff);
							sprintf (buff, "%s.o", tempfile);
							remove (buff);
							remove ("acs.err");

							InitGfx();
							CheckMouseDriver();
							RedrawMap = REDRAW_ALL;
						}
						chdir (cwd);
						free (cwd);
					}
				}
			}

			/* user wants to save the level data */
			else if ((IS_CTRL_KEY(key, 'Q') || (key & 0xFF00) == F2 || keychar == 'Q') && Registered)
			{
				char *outfile = GetWadFileName();
				if (outfile)
				{
					SaveLevelData( outfile);
					if (key == CTRL_Q)
					{
						SaveConfigFileOptions();
						break;
					}
				}
				RedrawMap = REDRAW_ALL;
				InfoDisplayed = FALSE;
			}
			
			/* user wants to save and change the episode and mission numbers */
			else if ((key & 0xFF00) == F3 && Registered)
			{
				char *outfile = NULL;
				MDirPtr newLevel, oldl, newl;
				char name[ 9];
				BCINT i;
				
				outfile = GetWadFileName();
				if (outfile)
				{
					SelectLevel();
					if (*LevelName)
					{
						/* horrible but it works... */
						sprintf( name, "%s", LevelName);
						newLevel = FindMasterDir( MasterDir, name);
						/* [RH] don't crash if the map doesn't exist yet */
						if (newLevel == NULL)
						{
							newLevel = AddMapLumps (MasterDir, name);
						}
						oldl = Level;
						newl = newLevel;
						for (i = 0; i < 12; i++)
						{
							newl->wadfile = oldl->wadfile;
							if (i > 0)
								newl->dir = oldl->dir;
							oldl = oldl->next;
							newl = newl->next;
						}
						Level = newLevel;
					}
					SaveLevelData( outfile);
				}
				RedrawMap = REDRAW_ALL;
			}
			
			/* user wants to get the 'Preferences' menu */
			else if (key == F5)
			{
				Preferences( -1, -1);
				RedrawMap = REDRAW_ALL;
			}

			/* user wants to get the 'Options' menu */
			else if (key == F7)
			{ 
				Options( -1, -1);
				RedrawMap = REDRAW_ALL;
			}

			else if (key == F4 && EditMode == OBJ_VERTEXES)
			{
				ShowDistance = !ShowDistance;
				RedrawMap = REDRAW_ALL;
			}

			else if (key == F4 && EditMode == OBJ_LINEDEFS)
			{ 
				AutoMap = !AutoMap;
				RedrawMap = REDRAW_ALL;
			}

			else if (key == F4 && EditMode == OBJ_THINGS)
			{
				ShowThings = !ShowThings;
				RedrawMap = REDRAW_ALL;
			}

			/* user wants to get the menu of misc. ops */
			else if ((key & 0xFF00) == F8)
			{
				if (Selected)
				{
					MiscOperations( -1, -1, EditMode, &Selected, CurObject);
				}
				else
				{
					if (CurObject >= 0)
					{
						SelectObject1( &Selected, CurObject);
					}
					MiscOperations( -1, -1, EditMode, &Selected, CurObject);
					if (CurObject >= 0)
					{
						UnSelectObject( &Selected, CurObject);
					}
				}
				CurObject = -1;
				RedrawMap = REDRAW_ALL;
				DraggingObject = FALSE;
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
			}
			
			/* user wants to insert a standard shape */
			else if ((key & 0xFF00) == F9)
			{
				BCINT savednum, i;
				Backup (~BTHINGS);
				strcpy (BackupText, "Insert Prefab Object");
				savednum = NumLineDefs;
				InsertStandardObject( -1, -1, MAPX( PointerX), MAPY( PointerY));
				if (NumLineDefs > savednum)
				{
					ForgetSelection( &Selected);
					EditMode = OBJ_LINEDEFS;
					for (i = savednum; i < NumLineDefs; i++)
					{
						SelectObject1( &Selected, i);
					}
					CurObject = NumLineDefs - 1;
					OldObject = -1;
					DraggingObject = FALSE;
					StretchSelBox = FALSE;
					GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
				}
				RedrawMap = REDRAW_ALL;
			}
			
			/* user wants to check his level */
			else if ((key & 0xFF00) == F10)
			{
				CheckLevel( -1, -1);
				RedrawMap = REDRAW_ALL;
			}

			/* user wants to change the edit mode */
			else if ((key & 0x00FF) == 0x0009 || (key & 0xFF00) == 0x0F00 ||
					 keychar == 'T' || keychar == 'V' || keychar == 'L' || keychar == 'S')
			{
				BCINT PrevMode = EditMode;
				SelPtr NewSel;
				
				if ((key & 0x00FF) == 0x0009) /* 'Tab' */
				{
					switch (EditMode)
					{
					case OBJ_THINGS:
						EditMode = OBJ_VERTEXES;
						break;
					case OBJ_VERTEXES:
						EditMode = OBJ_LINEDEFS;
						break;
					case OBJ_LINEDEFS:
						EditMode = OBJ_SECTORS;
						break;
					case OBJ_SECTORS:
						EditMode = OBJ_THINGS;
						break;
					}
				}
				else if ((key & 0xFF00) == 0x0F00) /* 'Shift-Tab' */
				{
					switch (EditMode)
					{
					case OBJ_THINGS:
						EditMode = OBJ_SECTORS;
						break;
					case OBJ_VERTEXES:
						EditMode = OBJ_THINGS;
						break;
					case OBJ_LINEDEFS:
						EditMode = OBJ_VERTEXES;
						break;
					case OBJ_SECTORS:
						EditMode = OBJ_LINEDEFS;
						break;
					}
				}
				else
				{
					if (keychar == 'T')
						EditMode = OBJ_THINGS;
					else if (keychar == 'V')
						EditMode = OBJ_VERTEXES;
					else if (keychar == 'L')
						EditMode = OBJ_LINEDEFS;
					else if (keychar == 'S')
						EditMode = OBJ_SECTORS;
					ForgetSelection( &Selected);
				}
				EdMode = EditMode;
				DrawMode = FALSE;
				
				/* special cases for the selection list... */
				if (Selected)
				{
					/* select all LineDefs bound to the selected Sectors */
					if (PrevMode == OBJ_SECTORS && EditMode == OBJ_LINEDEFS)
					{
						int l, sd;
						
						NewSel = NULL;
						for (l = 0; l < NumLineDefs; l++)
						{
							sd = LineDefs[ l].sidedef1;
							if (sd >= 0 && IsSelected( Selected, SideDefs[ sd].sector))
							{
								SelectObject1( &NewSel, l);
							}
							else
							{
								sd = LineDefs[ l].sidedef2;
								if (sd >= 0 && IsSelected( Selected, SideDefs[ sd].sector))
								{
									SelectObject1( &NewSel, l);
								}
							}
						}
						ForgetSelection( &Selected);
						Selected = NewSel;
					}
					/* select all Vertices bound to the selected LineDefs */
					else if (PrevMode == OBJ_LINEDEFS && EditMode == OBJ_VERTEXES)
					{
						NewSel = NULL;
						while (Selected)
						{
							if (!IsSelected( NewSel, LineDefs[ Selected->objnum].start))
							{
								SelectObject1( &NewSel, LineDefs[ Selected->objnum].start);
							}
							if (!IsSelected( NewSel, LineDefs[ Selected->objnum].end))
							{
								SelectObject1( &NewSel, LineDefs[ Selected->objnum].end);
							}
							UnSelectObject( &Selected, Selected->objnum); 
						}
						Selected = NewSel;
					}
					/* select all Sectors that have their LineDefs selected */
					else if (PrevMode == OBJ_LINEDEFS && EditMode == OBJ_SECTORS)
					{
						int l, sd;
						
						NewSel = NULL;
						/* select all Sectors... */
						for (l = 0; l < NumSectors; l++)
						{
							SelectObject1( &NewSel, l);
						}
						/* ... then unselect those that should not be in the list */
						for (l = 0; l < NumLineDefs; l++)
						{
							if (!IsSelected( Selected, l))
							{
								sd = LineDefs[ l].sidedef1;
								if (sd >= 0)
								{
									UnSelectObject( &NewSel, SideDefs[ sd].sector);
								}
								sd = LineDefs[ l].sidedef2;
								if (sd >= 0)
								{
									UnSelectObject( &NewSel, SideDefs[ sd].sector);
								}
							}
						}
						ForgetSelection( &Selected);
						Selected = NewSel;
					}
					/* select all LineDefs that have both ends selected */
					else if (PrevMode == OBJ_VERTEXES && EditMode == OBJ_LINEDEFS)
					{
						int l;
						
						NewSel = NULL;
						for (l = 0; l < NumLineDefs; l++)
						{
							if (IsSelected( Selected, LineDefs[ l].start) && IsSelected( Selected, LineDefs[ l].end))
							{
								SelectObject1( &NewSel, l);
							}
						}
						ForgetSelection( &Selected);
						Selected = NewSel;
					}
					/* unselect all */
					else
					{
						ForgetSelection( &Selected);
					}
				}
				CurObject = (GetMaxObjectNum( EditMode) >= 0 && Select0) ? 0 : -1;
				OldObject = -1;
				DraggingObject = FALSE;
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
				RedrawMap = REDRAW_ALL;
			}
			
			else if (keychar == 'U')
			{
				Undo();
				CurObject = -1;
				ForgetSelection( &Selected);
				RedrawMap = REDRAW_ALL;
			}

			else if (keychar == 'R')
			{
				DrawMode = DrawMode ? FALSE : TRUE;
				SetColor(DrawMode ? YELLOW : DARKGRAY);
				DrawScreenText(ScrMaxX - 220, 5, "DM");
			}


			/* user wants to display/hide the info box */
			else if (keychar == 'I')
			{
				if ((altkey & MOD_SHIFT) == 0x00)  /* no shift keys */
					InfoShown = !InfoShown;
				else
					DispSprite = !DispSprite;
				RedrawMap = REDRAW_ALL;
			}

			
			/* user wants to change the scale */
			else if ((keychar == '+' || keychar == '=') && Scale < 32.0)
			{ 
				double oscale = Scale;
#if 0
				if (Scale < 1.0)
					Scale = 1.0 / ((1.0 / Scale) - 1.0);
				else
					Scale = Scale * 2.0;
#else	// [RH] Zoom like the Build editor
				Scale *= 1.0625;
#endif
				OrigX += (PointerX - ScrCenterX) / oscale - (PointerX - ScrCenterX) / Scale;
				OrigY += (ScrCenterY - PointerY) / oscale - (ScrCenterY - PointerY) / Scale;
				RedrawMap = REDRAW_ALL;
			}
			else if ((keychar == '-' || keychar == '_') && Scale > 0.0125)
			{
				double oscale = Scale;
#if 0
				if (Scale < 1.0)
					Scale = 1.0 / ((1.0 / Scale) + 1.0);
				else
					Scale = Scale * 0.5;
#else	// [RH] Zoom like the Build editor
				Scale *= 0.9375;
#endif
				OrigX += (PointerX - ScrCenterX) / oscale - (PointerX - ScrCenterX) / Scale;
				OrigY += (ScrCenterY - PointerY) / oscale - (ScrCenterY - PointerY) / Scale;
				RedrawMap = REDRAW_ALL;
			}
			
			/* user wants to set the scale directly */
			else if (keychar >= '0' && keychar <= '9')
			{
				double oscale = Scale;
				Scale = (keychar == '0') ? 0.1 : 1.0 / (keychar - '0');
				OrigX += (PointerX - ScrCenterX) / oscale - (PointerX - ScrCenterX) / Scale;
				OrigY += (ScrCenterY - PointerY) / oscale - (ScrCenterY - PointerY) / Scale;
				RedrawMap = REDRAW_ALL;
			}
			
			/* user wants to move */
			else if ((key & 0xFF00) == CURS_UP)
			{
				if (circleline >= 0 && (altkey & MOD_SHIFT)) /* some shift key */
				{
					if (circlepoints < CIRCLEPOINTS_MAX)
					{
						if (lastcircleline >= 0)
						{
							SetColor (LIGHTGREEN);
							SetWriteMode (XOR_PUT);
							DrawArcSpots (circlex, circley, circlepoints);
							SetWriteMode (COPY_PUT);
							lastcircleline = -1;
						}
						++circlepoints;
					}
				}
				else if ((PointerY - MoveSpeed) >= 0)
				{
					if (UseMouse)
						SetMouseCoords( PointerX, PointerY - MoveSpeed);
					else
						PointerY -= MoveSpeed;
				}
			}
			else if ((key & 0xFF00) == CURS_DOWN)
			{
				if (circleline >= 0 && (altkey & MOD_SHIFT)) /* some shift key */
				{
					if (circlepoints > 1)
					{
						if (lastcircleline >= 0)
						{
							SetColor (LIGHTGREEN);
							SetWriteMode (XOR_PUT);
							DrawArcSpots (circlex, circley, circlepoints);
							SetWriteMode (COPY_PUT);
							lastcircleline = -1;
						}
						--circlepoints;
					}
				}
				else if ((PointerY + MoveSpeed) <= ScrMaxY)
				{
					if (UseMouse)
						SetMouseCoords( PointerX, PointerY + MoveSpeed);
					else
						PointerY += MoveSpeed;
				}
			}
			else if ((key & 0xFF00) == CURS_LEFT && (PointerX - MoveSpeed) >= 0)
			{
				if (UseMouse)
					SetMouseCoords( PointerX - MoveSpeed, PointerY);
				else
					PointerX -= MoveSpeed;
			}
			else if ((key & 0xFF00) == CURS_RIGHT && (PointerX + MoveSpeed) <= ScrMaxX)
			{
				if (UseMouse)
					SetMouseCoords( PointerX + MoveSpeed, PointerY);
				else
					PointerX += MoveSpeed;
			}
			
			/* user wants so scroll the map */
			else if ((key & 0xFF00) == PAGEUP && MAPY( ScrCenterY) < MapMaxY)
			{ 
				OrigY += (BCINT) (ScrCenterY / Scale);
				RedrawMap = REDRAW_ALL;
			}
			else if ((key & 0xFF00) == PAGEDOWN && MAPY( ScrCenterY) > MapMinY)
			{ 
				OrigY -= (BCINT) (ScrCenterY / Scale);
				RedrawMap = REDRAW_ALL;
			}
			else if ((key & 0xFF00) == HOME && MAPX( ScrCenterX) > MapMinX)
			{
				OrigX -= (BCINT) (ScrCenterX / Scale);
				RedrawMap = REDRAW_ALL;
			}
			else if ((key & 0xFF00) == END && MAPX( ScrCenterX) < MapMaxX)
			{
				OrigX += (BCINT) (ScrCenterX / Scale);
				RedrawMap = REDRAW_ALL;
			}
			
			/* user wants to change the movement speed */
			else if (keychar == ' ')
			{
				MoveSpeed = MoveSpeed == 2 ? DefaultLargeScroll : 2;
			}
			
			/* user wants to change the grid scale */
			else if (keychar == 'G')
			{
				if ((altkey & MOD_SHIFT) == 0x00) /* no shift keys */
				{
				
					if (GridScale == 0)
						GridScale = MaxGrid;
					else if (GridScale > MinGrid)
						GridScale /= 2;
					else
						GridScale = 0;
				}
				else
				{
					if (GridScale == 0)
						GridScale = MinGrid;
					else if (GridScale < MaxGrid)
						GridScale *= 2;
					else
						GridScale = 0;
				}
				RedrawMap = REDRAW_ALL;
			}
			else if (keychar == 'K')
			{
				if ((altkey & MOD_SHIFT) == 0x00)  /* no shift keys */
				{
					GridScale = 0;
					RedrawMap = REDRAW_ALL;
				}
			}
			else if (keychar == 'H')
			{
				if ((altkey & MOD_SHIFT) != 0x00)  /* shift key pressed */
					GridScale = 0;
				else
					GridShown = !GridShown;
				RedrawMap = REDRAW_ALL;
			}


			/* user wants to toggle drag mode */
			else if (keychar == 'D')
			{
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
				if (DraggingObject)
				{
					DraggingObject = FALSE;
					if (EditMode == OBJ_VERTEXES)
					{
						if (Selected == NULL && CurObject >= 0)
						{
							SelectObject1( &Selected, CurObject);
							if (AutoMergeVertices( &Selected))
							{
								ForgetSelection( &Selected);
								RedrawMap = REDRAW_ALL;
							}
							ForgetSelection( &Selected);
						}
						if (AutoMergeVertices( &Selected))
						{
							ForgetSelection( &Selected);
							RedrawMap = REDRAW_ALL;
						}
					}
					else if (EditMode == OBJ_LINEDEFS)
					{
						SelPtr NewSel, cur;
						
						NewSel = NULL;
						if (Selected == NULL && CurObject >= 0)
						{
							SelectObject1( &NewSel, LineDefs[ CurObject].start);
							SelectObject1( &NewSel, LineDefs[ CurObject].end);
						}
						else
						{
							for (cur = Selected; cur; cur = cur->next)
							{
								if (!IsSelected( NewSel, LineDefs[ cur->objnum].start))
								{
									SelectObject1( &NewSel, LineDefs[ cur->objnum].start);
								}
								if (!IsSelected( NewSel, LineDefs[ cur->objnum].end))
								{
									SelectObject1( &NewSel, LineDefs[ cur->objnum].end);
								}
							}
						}
						if (AutoMergeVertices( &NewSel))
							RedrawMap = REDRAW_ALL;
						ForgetSelection( &NewSel);
					}
				}
				else
				{
					DraggingObject = TRUE;
					if (EditMode == OBJ_THINGS && CurObject >= 0)
					{
				    	Backup (BTHINGS);
						strcpy (BackupText, "Drag Things");
						MoveObjectsToCoords( EditMode, NULL, Things[ CurObject].xpos, Things[ CurObject].ypos, 0);
					}
					else if (EditMode == OBJ_VERTEXES && CurObject >= 0)
					{
				    	Backup (~BTHINGS);
						strcpy (BackupText, "Drag Vertexes");
						MoveObjectsToCoords( EditMode, NULL, Vertexes[ CurObject].x, Vertexes[ CurObject].y, 0);
					}
					else
					{
				    	Backup (~BTHINGS);
						strcpy (BackupText, "Drag Sectors");
						if (EditMode == OBJ_LINEDEFS) 
								strcpy (BackupText, "Drag LineDefs");
						MoveObjectsToCoords( EditMode, NULL, MAPX( PointerX), MAPY( PointerY), GridScale);
					}
				}
			}
			
			/* user wants to select the next or previous object */
			else if (keychar == 'N' && circleline < 0)
			{
				if (CurObject < GetMaxObjectNum(EditMode))
				{
					CurObject++;
					if (Selected)
					{
						while (!IsSelected(Selected,CurObject))
							CurObject = CurObject<GetMaxObjectNum(EditMode)? CurObject+1 : 0;
						GoToObject(EditMode,CurObject);
					}
				}
				else if (GetMaxObjectNum(EditMode) >= 0)
				{
					CurObject = 0;
					if (Selected)
					{
						while (!IsSelected(Selected,CurObject))
							CurObject = CurObject<GetMaxObjectNum(EditMode)? CurObject+1 : 0;
						GoToObject(EditMode,CurObject);
					}
				}
/*
				if (CurObject < GetMaxObjectNum( EditMode))
					CurObject++;
				else if (GetMaxObjectNum( EditMode) >= 0)
					CurObject = 0;
				else
					CurObject = -1;
*/
				RedrawMap = REDRAW_ALL;
			}
			else if (keychar == 'P' && circleline < 0)
			{
				if (CurObject > 0)
				{
					CurObject--;
					if (Selected)
					{
						while (!IsSelected(Selected,CurObject))
							CurObject = CurObject>0? CurObject-1 : GetMaxObjectNum(EditMode);
						GoToObject(EditMode,CurObject);
					}
				}
				else
				{
					CurObject = GetMaxObjectNum(EditMode);
					if (Selected)
					{
						while (!IsSelected(Selected,CurObject))
							CurObject = CurObject>0? CurObject-1 : GetMaxObjectNum(EditMode);
						GoToObject(EditMode,CurObject);
					}
				}
/*
				if (CurObject > 0)
					CurObject--;
				else
					CurObject = GetMaxObjectNum( EditMode);
*/
				RedrawMap = REDRAW_ALL;
			}
			else if ((keychar == 'J' || keychar == '#') && circleline < 0)
			{
				OldObject = InputObjectNumber( -1, -1, EditMode, CurObject);
				if (OldObject >= 0)
				{
					CurObject = OldObject;
					GoToObject( EditMode, CurObject);
				}
				else
				{
					OldObject = CurObject;
				}
				RedrawMap = REDRAW_ALL;
			}
			else if (keychar == 'B' && EditMode == OBJ_THINGS)
			{
				val = InputIntegerValue( -1, -1, 0, 32767, 1);
				if (val >= 1)
				{
					val = FindThingByTID (0, val);
					if (val >= 0)
					{
						GoToObject (EditMode, val);
						RedrawMap = REDRAW_ALL;
					}
				}
			}

			/* user wants to mark/unmark an object or a group of objects */
			else if (keychar == 'M' && circleline < 0)
			{
				if (StretchSelBox)
				{
					SelPtr oldsel;
					
					/* select all objects in the selection box */
					StretchSelBox = FALSE;
					GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
					RedrawMap = REDRAW_ALL;
					/* additive selection box or not? */
					if (!AdditiveSelBox)
						ForgetSelection( &Selected);
					else
						oldsel = Selected;
					Selected = SelectObjectsInBox( EditMode, SelBoxX, SelBoxY, MAPX( PointerX), MAPY( PointerY));
					if (AdditiveSelBox)
					{
						while (oldsel != NULL)
						{
							if (! IsSelected( Selected, oldsel->objnum))
							{
								SelectObject1( &Selected, oldsel->objnum);
							}
							UnSelectObject( &oldsel, oldsel->objnum);
						}
					}
					if (Selected)
					{
						CurObject = Selected->objnum;
						PlaySound( 440, 10);
					}
					else
					{
						CurObject = -1;
					}
				}
				else if ((altkey & MOD_SHIFT) == 0x00) /* no shift keys */
				{
					if (CurObject >= 0)
					{
						/* mark or unmark one object */
						if (IsSelected( Selected, CurObject))
						{
							UnSelectObject( &Selected, CurObject);
						}
						else
						{
							SelectObject1( &Selected, CurObject);
						}
						HighlightObject( EditMode, CurObject, GREEN);
						if (Selected)
							PlaySound( 440, 10);
						DraggingObject = FALSE;
					}
					else
					{
						DEU_Beep();
					}
				}
				else
				{
					/* begin "stretch selection box" mode */
					SelBoxX = MAPX( PointerX);
					SelBoxY = MAPY( PointerY);
					StretchSelBox = TRUE;
					GrMouseSetColors(TranslateToGameColor(YELLOW),GrNOCOLOR);
					DraggingObject = FALSE;
				}
				GetCount(Selected, 0);
				DrawInfoBar(EditMode, GridScale);
			}
			
			/* user wants to clear all marks and redraw the map */
			else if (keychar == 'C')
			{
				ForgetSelection( &Selected);
				RedrawMap = REDRAW_ALL;
				DraggingObject = FALSE;
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
				DrawMode = FALSE;
			}
			
			/* user wants to copy a group of objects */
			else if (keychar == 'O' && CurObject >= 0 && circleline < 0)
			{
				/* copy the object(s) */
				if (Selected == NULL)
				{
					SelectObject1( &Selected, CurObject);
				}
				if (EditMode == OBJ_THINGS)
				{
					Backup (BTHINGS);
					strcpy (BackupText, "Copy Things");
				}
				else if (EditMode == OBJ_VERTEXES)
				{
					Backup (BVERTEXES);
					strcpy (BackupText, "Copy Vertexes");
				}
				else if (EditMode == OBJ_LINEDEFS)
				{
					Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS);
					strcpy (BackupText, "Copy LineDefs");
				}
				else if (EditMode == OBJ_SECTORS)
				{
					Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS | BSECTORS);
					strcpy (BackupText, "Copy Sectors");
				}
				CopyObjects( EditMode, Selected);
				/* enter drag mode */
				DraggingObject = TRUE;
				CurObject = Selected->objnum;
				if (EditMode == OBJ_THINGS)
					MoveObjectsToCoords( EditMode, NULL, Things[ CurObject].xpos, Things[ CurObject].ypos, 0);
				else if (EditMode == OBJ_VERTEXES) 
					MoveObjectsToCoords( EditMode, NULL, Vertexes[ CurObject].x, Vertexes[ CurObject].y, 0);
				else
					MoveObjectsToCoords( EditMode, NULL, MAPX( PointerX), MAPY( PointerY), GridScale);
				RedrawMap = REDRAW_ALL;
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
			}

			else if ((key == 0x0221))		// !
			{ 
				TermGfx();
				printf("Type Exit to return\n");
				system("");
				InitGfx();
				CheckMouseDriver();
				RedrawMap = REDRAW_ALL;
			}

/************/
/* HOT KEYS */
/************/

			else if (IS_CTRL_KEY(key, 'S') && EditMode == OBJ_THINGS)
				RedrawMap = FunctionThingSpecial(CurObject, Selected);

			else if (IS_CTRL_KEY(key, 'T') && EditMode == OBJ_THINGS)
				RedrawMap = FunctionThingType(CurObject, Selected);

			else if (IS_CTRL_KEY(key, 'T') && EditMode == OBJ_SECTORS)
				RedrawMap = FunctionSectorType(CurObject, Selected);

			else if ((IS_CTRL_KEY(key, 'J') || IS_CTRL_KEY(key, 'D')) && EditMode == OBJ_VERTEXES)
			{
				RedrawMap = FunctionVertexDeleteJoin(CurObject, Selected);
				ForgetSelection ( &Selected);
				CurObject = -1;
			}

			else if (IS_CTRL_KEY(key, 'T') && EditMode == OBJ_LINEDEFS)
				RedrawMap = FunctionLineDefSpecial(CurObject, Selected);

			else if ((IS_CTRL_KEY(key, 'J') || IS_CTRL_KEY(key, 'D')) && EditMode == OBJ_LINEDEFS)
			{ 
				RedrawMap = FunctionLineDefDeleteJoin(CurObject, Selected);
				ForgetSelection ( &Selected);
			}

			else if (IS_CTRL_KEY(key, 'V') && EditMode == OBJ_LINEDEFS)
			{ 
				RedrawMap = FunctionVertexSplitLineDef(CurObject, Selected);
				ForgetSelection ( &Selected);
			}
			
			else if (IS_CTRL_KEY(key, 'F') && EditMode == OBJ_LINEDEFS)
				RedrawMap = FunctionLineDefFlipSideDefs(CurObject, Selected);

			else if (IS_CTRL_KEY(key, 'S') && EditMode == OBJ_LINEDEFS)
				RedrawMap = FunctionLineDefSwapSideDefs(CurObject, Selected);

			else if (IS_CTRL_KEY(key, 'M') && key != ENTER && EditMode == OBJ_VERTEXES)
			{ 
				RedrawMap = FunctionVertexMerge(CurObject, &Selected);
				ForgetSelection( &Selected);
			} 

			else if (IS_CTRL_KEY(key, 'X') && EditMode == OBJ_LINEDEFS)
			{ 
				Backup (BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Align Textures on X without Checking");
				AlignTexturesX ( &Selected, TRUE, FALSE);
				RedrawMap = REDRAW_ALL; 
			}

			else if (keychar == 'X' && EditMode == OBJ_LINEDEFS)
			{
				Backup (BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Align Textures on X with Checking");
				AlignTexturesX ( &Selected, FALSE, FALSE);
				RedrawMap = REDRAW_ALL; 
			}

			else if (IS_CTRL_KEY(key, 'Y') && EditMode == OBJ_LINEDEFS)
			{
				Backup (BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Align Textures on Y without Checking");
				AlignTexturesY ( &Selected, TRUE, FALSE);
				RedrawMap = REDRAW_ALL; 
			}

			else if (keychar == 'Y' && EditMode == OBJ_LINEDEFS)
			{
				Backup (BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Align Textures on Y with Checking");
				AlignTexturesY ( &Selected, FALSE, FALSE);
				RedrawMap = REDRAW_ALL; 
			}

			else if (IS_CTRL_KEY(key, 'M') && key != ENTER && EditMode == OBJ_LINEDEFS)
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 0);

			else if (IS_CTRL_KEY(key, 'P') && EditMode == OBJ_LINEDEFS)
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 1);

			else if (key == CTRL_2 && EditMode == OBJ_LINEDEFS) 
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 2);

			else if (key == CTRL_6) 
				SaveConfigFileOptions();

			else if (IS_CTRL_KEY(key, 'U') && EditMode == OBJ_LINEDEFS)
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 3);

			else if (IS_CTRL_KEY(key, 'L') && EditMode == OBJ_LINEDEFS) 
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 4);

			else if (IS_CTRL_KEY(key, 'R') && EditMode == OBJ_LINEDEFS) 
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 5);

			else if (IS_CTRL_KEY(key, 'B') && EditMode == OBJ_LINEDEFS) 
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 6);

			else if (IS_CTRL_KEY(key, 'H') && EditMode == OBJ_LINEDEFS) 
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 7);

			else if (IS_CTRL_KEY(key, 'O') && EditMode == OBJ_LINEDEFS) 
				RedrawMap = FunctionLineDefFlag(CurObject, Selected, 8);

			else if (IS_CTRL_KEY(key, 'G') && EditMode == OBJ_LINEDEFS)
			{ 
				SelPtr cur;
//				BCINT x0, y0;
				Backup (BLINEDEFS);
				strcpy (BackupText, "Reset LineDef Arguments");
				for(cur = Selected; cur; cur = cur->next)
				{
					if(cur->objnum != CurObject)
						memset (LineDefs[ cur->objnum].arg, 0, sizeof(LineDefs[cur->objnum].arg));
				}
				memset (LineDefs[CurObject].arg, 0, sizeof(LineDefs[CurObject].arg));
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL; 
			}


			else if (IS_CTRL_KEY(key, 'G') && EditMode == OBJ_THINGS)
			{
				SelPtr cur;
//				BCINT x0, y0;
				Backup (BTHINGS);
				strcpy (BackupText, "Reset Thing Arguments");
				for (cur = Selected; cur; cur = cur->next)
				{
					if (cur->objnum != CurObject)
						memset (Things[cur->objnum].arg, 0, sizeof(Things[cur->objnum].arg));
				}
				memset (Things[CurObject].arg, 0, sizeof(Things[CurObject].arg));
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL; 
			}

			else if (IS_CTRL_KEY(key, 'G') && EditMode == OBJ_SECTORS)
			{
				SelPtr cur;
				BCINT x0, y0;
				Backup (BSECTORS);
				strcpy (BackupText, "Change Sector LineDef Tag");
				x0 = ((ScrMaxX - 280) /2);
				y0 = ((ScrMaxY / 2 ) + 29);
				DrawScreenBox3D( x0, y0 , x0 + 25 + (8 * 32), y0 + 55);
				DrawScreenText( x0 + 10, y0 + 10, "Current Tag: %d",Sectors[CurObject].tag);
				DrawScreenText( -1, y0 + 24, "Next Free Tag: %d",FindFreeTag());
				DrawScreenText( -1, y0 + 38, "Last Highlighted LineDef Tag: %d",LastLineDefTag);
				if (Sectors[ CurObject].tag == 0)	/* jff allow 999 tags */
					val = InputIntegerValue( -1, -1, 0, 32767, FindFreeTag());
				else 
					val = InputIntegerValue( -1, -1, 0, 32767, Sectors[ CurObject].tag);
				if (val >= 0) {
					for(cur = Selected; cur; cur = cur->next)
					{
						if(cur->objnum != CurObject) 
							Sectors[ cur->objnum].tag = val;
					}
					Sectors[ CurObject].tag = val;
					LastSectorTag = Sectors[CurObject].tag;
					MadeChanges = TRUE;
				}
				RedrawMap = REDRAW_ALL; 
			}

			else if (IS_CTRL_KEY(key, 'N') && EditMode == OBJ_LINEDEFS)
			{
				SelPtr cur;
				Backup (BLINEDEFS);
				strcpy (BackupText, "Set LineDef Type To Normal And Tag To 0");
				for (cur = Selected; cur; cur = cur->next)
				{
					if(cur->objnum != CurObject)
					{
						LineDefs[ cur->objnum].special = 0;
						memset (LineDefs[ cur->objnum].arg, 0, sizeof(LineDefs[cur->objnum].arg));
					}
				}
				LineDefs[ CurObject].special = 0;
				memset (LineDefs[ CurObject].arg, 0, sizeof(LineDefs[CurObject].arg));
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}

			else if (IS_CTRL_KEY(key, 'N') && EditMode == OBJ_SECTORS)
			{
				SelPtr cur;
				Backup (BSECTORS);
				strcpy (BackupText, "Set Sector Type To Normal And Tag To 0");
				for (cur = Selected; cur; cur = cur->next)
				{
					if(cur->objnum != CurObject)
					{
						Sectors[ cur->objnum].special = 0;
						Sectors[ cur->objnum].tag = 0;
					}
				}
				Sectors[ CurObject].special = 0;
				Sectors[ CurObject].tag = 0;
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}


			else if (IS_CTRL_KEY(key, 'F') && EditMode == OBJ_THINGS)
				RedrawMap = FunctionThingFlags(CurObject, Selected);

			else if (IS_CTRL_KEY(key, 'C') && EditMode == OBJ_LINEDEFS)
				RedrawMap = FunctionLineDefFlags(CurObject, Selected);

			else if (IS_CTRL_KEY(key, 'E') && EditMode == OBJ_LINEDEFS) 
				RedrawMap = FunctionLineDefTrigger(CurObject, Selected);

			else if (IS_CTRL_KEY(key, 'B') && EditMode == OBJ_THINGS)
				RedrawMap = FunctionThingFlag(CurObject, Selected, 5);

			else if (IS_CTRL_KEY(key, 'D') && EditMode == OBJ_THINGS)
				RedrawMap = FunctionThingFlag(CurObject, Selected, 4);

			else if (IS_CTRL_KEY(key, 'A') && EditMode == OBJ_LINEDEFS)
			{
				RedrawMap = FunctionLineDefSplitAdd(CurObject, Selected);
				ForgetSelection( &Selected);
			}

			else if (IS_CTRL_KEY(key, 'K') && EditMode == OBJ_LINEDEFS)
			{
				RedrawMap = FunctionConnectLineDefsSplitDonut(CurObject, Selected);
				ForgetSelection( &Selected);
			}

			else if (IS_CTRL_KEY(key, 'A') && EditMode == OBJ_VERTEXES)
			{
				RedrawMap = FunctionVertexSplitSector(CurObject, Selected);
				ForgetSelection( &Selected);
			}

			/* user wants to rotate things */
			else if(EditMode == OBJ_THINGS && ((keychar == ',') || (keychar == '.')))
			{
				BCINT temp, rot = (keychar == ',') ? 45 : 315;
				SelPtr cur;
				Backup (BTHINGS);
				strcpy (BackupText, "Change Thing Angle");
				for(cur = Selected; cur; cur = cur->next)
				{
					if(cur->objnum != CurObject)
					{
						temp = (Things[cur->objnum].angle + rot) % 360;
						Things[cur->objnum].angle = temp;
					}
				}
				temp = (Things[CurObject].angle + rot) % 360;
				Things[CurObject].angle = temp;
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}

			/* user wants to change thing's z position */
			else if(EditMode == OBJ_THINGS && ((keychar == '<') || (keychar == '>')))
			{
				BCINT temp, val = (keychar == '<') ? -16 : 16;
				SelPtr cur;
				Backup (BTHINGS);
				strcpy (BackupText, "Change Thing Height");
				for(cur = Selected; cur; cur = cur->next)
				{
					if(cur->objnum != CurObject)
					{
						temp = (Things[cur->objnum].zpos + val);
						Things[cur->objnum].zpos = temp;
					}
				}
				temp = (Things[CurObject].zpos + val);
				Things[CurObject].zpos = temp;
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}

			else if(EditMode == OBJ_SECTORS && IS_CTRL_KEY(key, 'F'))
				RedrawMap = FunctionSectorFloorHeight(CurObject, Selected);

			else if(EditMode == OBJ_SECTORS && IS_CTRL_KEY(key, 'C')) 
				RedrawMap = FunctionSectorCeilingHeight(CurObject, Selected);

			else if(EditMode == OBJ_SECTORS && IS_CTRL_KEY(key, 'L'))
				RedrawMap = FunctionSectorLightLevel(CurObject, Selected);
			

			else if(keychar == '[' || keychar == ']')
			{
				BCINT angle = (keychar == '[') ? 90 : 270;
				Backup (EditMode == OBJ_THINGS ? BTHINGS : BVERTEXES);
				strcpy (BackupText, "Rotate Objects");
				RotateAndScaleObjects(EditMode, Selected, (double) angle * 0.0174533, (double) 100 * 0.01, 0);
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}
/*
			else if(keychar == '{' || keychar == '}') {
				BCINT temp, scale = (keychar == '}') ? 50 : 200;
				Backup (BTHINGS | BSECTORS | BVERTEXES | BLINEDEFS | BLINEDEFS);
				strcpy (BackupText, "Scale Objects");
				RotateAndScaleObjects(EditMode, Selected, (double) 0 * 0.0174533, (double) scale * 0.01, 0);
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}

*/
			else if(keychar == '/')
			{
				CenterMapAroundCoords((MapMinX + MapMaxX) / 2, (MapMinY + MapMaxY) / 2);
				RedrawMap = REDRAW_ALL;
			}

			else if(keychar == '?')
			{
				ShowAllMap ();
				RedrawMap = REDRAW_ALL;
			}

			else if(EditMode == OBJ_SECTORS && ((keychar == '<') || (keychar == '>')))
			{
				BCINT temp, amount = (keychar == '>') ? Increment : Increment - (2 * Increment);
				SelPtr cur;
				Backup (BSECTORS);
				strcpy (BackupText, "Change Sector Ceiling Heights");
				for(cur = Selected; cur; cur = cur->next)
				{
					if(cur->objnum != CurObject)
					{
						temp = (Sectors[cur->objnum].ceilh + amount);
						Sectors[cur->objnum].ceilh = temp;
					}
				}
				temp = (Sectors[CurObject].ceilh + amount);
				Sectors[CurObject].ceilh = temp;
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}


			else if(EditMode == OBJ_SECTORS && ((keychar == ',') || (keychar == '.')))
			{
				BCINT temp, amount = (keychar == '.') ? Increment : Increment - (2 * Increment);
				SelPtr cur;
				Backup (BSECTORS);
				strcpy (BackupText, "Change Sector Floor Heights");
				for(cur = Selected; cur; cur = cur->next)
				{
					if(cur->objnum != CurObject)
					{
						temp = (Sectors[cur->objnum].floorh + amount);
						Sectors[cur->objnum].floorh = temp;
					}
				}
				temp = (Sectors[CurObject].floorh + amount);
				Sectors[CurObject].floorh = temp;
				MadeChanges = TRUE;
				RedrawMap = REDRAW_ALL;
			}

			/* user wants to edit the current object */
			else if ((key & 0x00ff) == 0x0D && CurObject >= 0)
			{
				if (Selected)
				{
					EditObjectsInfo( 0, 30, EditMode, Selected);
				}
				else
				{
					SelectObject1( &Selected, CurObject);
					EditObjectsInfo( 0, 30, EditMode, Selected);
					UnSelectObject( &Selected, CurObject);
				}
				RedrawMap = REDRAW_ALL;
				DraggingObject = FALSE;
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
			}
			
			/* user wants to delete the current object */
			else if ((key & 0xFF00) == DELETE && CurObject >= 0 && circleline < 0)
			{
				if (EditMode == OBJ_THINGS || Expert || Confirm( -1, -1,
						(Selected ? "Do You Really Want To Delete These Objects?" : "Do You Really Want To Delete This Object?"),
						(Selected ? "This Will Also Delete The Objects Bound To Them." : "This Will Also Delete The Objects Bound To It.")))
				{
					if (EditMode == OBJ_THINGS)
					{
						Backup (BTHINGS);
						strcpy (BackupText, "Delete Things");
					}
					else if (EditMode == OBJ_VERTEXES)
					{
						Backup (~BTHINGS);
						strcpy (BackupText, "Delete Vertexes");
					}
					else if (EditMode == OBJ_LINEDEFS)
					{
						Backup (~BTHINGS);
						strcpy (BackupText, "Delete LineDefs");
					}
					else if (EditMode == OBJ_SECTORS)
					{
						Backup (~BTHINGS);
						strcpy (BackupText, "Delete Sectors");
					}
					if (Selected)
					{
	    					/* SO 23/7/95:
						This is to fix a bug which has been present
						in DEU as far back as I have the source for! */
						SortSelection(&Selected);
						DeleteObjects( EditMode, &Selected);
					}
					else
					{
						DeleteObject1( EditMode, CurObject);
					}
					CurObject = -1;
				}
				DraggingObject = FALSE;
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
				RedrawMap = REDRAW_ALL;
			}

			/* user wants to insert a new object */
			else if ((key & 0xFF00) == INSERT)
			{
				SelPtr cur;

				if (circleline >= 0)
				{
					Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS);
					strcpy (BackupText, "Create an Arc From a LineDef");
				}
				else if (EditMode == OBJ_THINGS)
				{
					Backup (BTHINGS);
					strcpy (BackupText, "Insert Thing");
				}
				else if (EditMode == OBJ_VERTEXES)
				{
					Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS);	//jff change duplicate
					strcpy (BackupText, "Insert Vertex");		//BLINEDEFS to BSIDEDEFS
				}
				else if (EditMode == OBJ_LINEDEFS)
				{
					Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS);
					strcpy (BackupText, "Insert LineDef(s)");
				}
				else if (EditMode == OBJ_SECTORS)
				{
					Backup (BVERTEXES | BSIDEDEFS | BLINEDEFS);
					strcpy (BackupText, "Insert Sector");
				}

				/* [RH] new special case: convert a single linedef to an arc */
				if (circleline >= 0)
				{
					SelPtr sel;
					int end = LineDefs[circleline].end;
					int i, line;

					if (!(sel = SelectPointer (Selected, circleline)))
					{
						sel = SelectObject1 (&Selected, circleline);
					}
					line = circleline;
					for (i = 0; i < circlepoints; ++i)
					{
						InsertObject (OBJ_VERTEXES, -1, circlex[i+1], circley[i+1]);
						LineDefs[line].end = NumVertexes - 1;
						InsertObject (OBJ_LINEDEFS, circleline, 0, 0);
						line = NumLineDefs - 1;
						LineDefs[line].start = NumVertexes - 1;
						if (LineDefs[circleline].sidedef1 >= 0)
						{
							InsertObject (OBJ_SIDEDEFS, LineDefs[circleline].sidedef1, 0, 0);
							LineDefs[line].sidedef1 = NumSideDefs - 1;
						}
						if (LineDefs[circleline].sidedef2 >= 0)
						{
							InsertObject (OBJ_SIDEDEFS, LineDefs[circleline].sidedef2, 0, 0);
							LineDefs[line].sidedef2 = NumSideDefs - 1;
						}
						sel = SelectObjectAfter (sel, line);
					}
					LineDefs[line].end = end;
					circleline = -1;
				}

				/* first special case: if several Vertices are selected, add new LineDefs */
				else if (EditMode == OBJ_VERTEXES && Selected != NULL && Selected->next != NULL)
				{
					BCINT firstv;
					
					if (Selected->next->next != NULL)
						firstv = Selected->objnum;
					else
						firstv = -1;
					EditMode = OBJ_LINEDEFS;
					strcpy (BackupText, "Insert LineDef(s) Between Vertexes");
					/* create LineDefs between the Vertices */
					for (cur = Selected; cur->next; cur = cur->next)
					{
						/* check if there is already a LineDef between the two Vertices */
						for (CurObject = 0; CurObject < NumLineDefs; CurObject++)
						{
							if ((LineDefs[ CurObject].start == cur->next->objnum && LineDefs[ CurObject].end == cur->objnum)
								|| (LineDefs[ CurObject].end == cur->next->objnum && LineDefs[ CurObject].start == cur->objnum))
							{
								break;
							}
						}
						if (CurObject < NumLineDefs)
						{
							cur->objnum = CurObject;
						}
						else
						{
							InsertObject( OBJ_LINEDEFS, -1, 0, 0);
							CurObject = NumLineDefs - 1;
							LineDefs[ CurObject].end = cur->next->objnum;
							LineDefs[ CurObject].start = cur->objnum;
							cur->objnum = CurObject;
						}
					}
					/* close the polygon if there are more than 2 Vertices */
					if (firstv >= 0 && (altkey & MOD_SHIFT) != 0x00)  /* shift key pressed */
					{
						for (CurObject = 0; CurObject < NumLineDefs; CurObject++)
							if ((LineDefs[ CurObject].start == firstv && LineDefs[ CurObject].end == cur->objnum)
								|| (LineDefs[ CurObject].end == firstv && LineDefs[ CurObject].start == cur->objnum))
								break;
						if (CurObject < NumLineDefs)
							cur->objnum = CurObject;
						else
						{
							InsertObject( OBJ_LINEDEFS, -1, 0, 0);
							CurObject = NumLineDefs - 1;
							LineDefs[ CurObject].end = firstv;
							LineDefs[ CurObject].start = cur->objnum;
							cur->objnum = CurObject;
						}
					}
					else
					{
						UnSelectObject( &Selected, cur->objnum);
					}
				}
				/* second special case: if several LineDefs are selected, add new SideDefs and one Sector */
				else if (EditMode == OBJ_LINEDEFS && Selected != NULL)
				{
					for (cur = Selected; cur; cur = cur->next)
					{
						if (LineDefs[ cur->objnum].sidedef1 >= 0 && LineDefs[ cur->objnum].sidedef2 >= 0)
						{
							char msg[ 80];
							
							DEU_Beep();
							sprintf( msg, "LineDef #%d Already Has Two SideDefs", cur->objnum);
							Notify( -1, -1, "Error: Cannot Add The New Sector", msg);
							break;
						}
					}
					if (cur == NULL)
					{
						EditMode = OBJ_SECTORS;
						strcpy (BackupText, "Insert Sector");
						InsertObject( OBJ_SECTORS, -1, 0, 0);
						CurObject = NumSectors - 1;
						for (cur = Selected; cur; cur = cur->next)
						{
							InsertObject( OBJ_SIDEDEFS, -1, 0, 0);
							SideDefs[ NumSideDefs - 1].sector = CurObject;
							if (LineDefs[ cur->objnum].sidedef1 >= 0)
							{
								BCINT s;
								
								s = SideDefs[ LineDefs[ cur->objnum].sidedef1].sector;
								if (s >= 0)
								{
									Sectors[ CurObject].floorh = Sectors[ s].floorh;
									Sectors[ CurObject].ceilh = Sectors[ s].ceilh;
									strncpy( Sectors[ CurObject].floort, Sectors[ s].floort, 8);
									strncpy( Sectors[ CurObject].ceilt, Sectors[ s].ceilt, 8);
									Sectors[ CurObject].light = Sectors[ s].light;
								}
								LineDefs[ cur->objnum].sidedef2 = NumSideDefs - 1;
								LineDefs[ cur->objnum].flags = 4;
								strncpy( SideDefs[ NumSideDefs - 1].tex3, "-", 8);
								strncpy( SideDefs[ LineDefs[ cur->objnum].sidedef1].tex3, "-", 8);
							}
							else
							{
								LineDefs[ cur->objnum].sidedef1 = NumSideDefs - 1;
							}
						}
						ForgetSelection( &Selected);
						SelectObject1( &Selected, CurObject);
					}
				}
				/* normal case: add a new object of the current type */
				else
				{
					ForgetSelection( &Selected);
					if (GridScale > 0)
						InsertObject( EditMode, CurObject, (BCINT) ((int)(MAPX( PointerX) + GridScale / 2)) & ((int)(~(GridScale - 1))), (BCINT) ((int)(MAPY( PointerY) + GridScale / 2)) & (int)(~(GridScale - 1)));
					else
						InsertObject( EditMode, CurObject, MAPX( PointerX), MAPY( PointerY));
					CurObject = GetMaxObjectNum( EditMode);
					if (EditMode == OBJ_LINEDEFS)
					{
						if (! Input2VertexNumbers( -1, -1, "Choose The Two Vertices For The New LineDef",
												  &(LineDefs[ CurObject].start), &(LineDefs[ CurObject].end)))
						{
							DeleteObject1( EditMode, CurObject);
							CurObject = -1;
						}
					}
					else if (EditMode == OBJ_VERTEXES)
					{
						if (CurObject>-1)
						{
							SelectObject1( &Selected, CurObject);
							if (AutoMergeVertices( &Selected))
								RedrawMap = REDRAW_ALL;
							ForgetSelection( &Selected);
						}
					}
				}
				DraggingObject = FALSE;
				StretchSelBox = FALSE;
				GrMouseSetColors(TranslateToGameColor(WHITE),GrNOCOLOR);
				RedrawMap = REDRAW_ALL;
			}

			/* user likes music */
			else
			{
				DEU_Beep();
			}

			
			/* redraw the (keyboard) pointer */
			if (FakeCursor)
			{
				HideMousePointer();
				DrawPointer();
				ShowMousePointer();
			}
		}
		
		/* check if Scroll Lock is off */
		if ((bioskey (2) & MOD_SCROLL) == 0x00 && MoveSpeed > 0)
		{
			BCINT moveX = 0, moveY = 0;

			/* move the map if the pointer is near the edge of the screen */
			if (PointerY <= (UseMouse ? 0 : 20))
			{
				if (! UseMouse)
					PointerY += MoveSpeed;
				if (OrigY < 32767)
					moveY = 1;
			}
			else if (PointerY >= ScrMaxY - (UseMouse ? 0 : 20))
			{
				if (! UseMouse)
					PointerY -= MoveSpeed;
				if (OrigY > -32768)
					moveY = -1;
			}
			if (PointerX <= (UseMouse ? 0 : 20))
			{
				if (! UseMouse)   
					PointerX += MoveSpeed;
				if (OrigX > -32768)
					moveX = -1;
			}
			else if (PointerX >= ScrMaxX - (UseMouse ? 0 : 20))
			{
				if (! UseMouse)
					PointerX -= MoveSpeed;
				if (OrigX < 32767)
					moveX = 1;
			}

			if (moveX | moveY)
			{
				BCINT rate;
				int diff;

				// [RH] Make scroll rate constant for varying level complexities
#ifdef __WATCOMC__
				clock_t now;

				now = clock ();
				if (lastscroll == 0)
				{
					diff = 1000000/20;
				}
				else
				{
					diff = (now - lastscroll)*(1000000/CLOCKS_PER_SEC);
				}
#else
				struct timeval now;
				if (SmoothMouseScroll)
				{
					uclock_t clocks = uclock();
					now.tv_sec = clocks / UCLOCKS_PER_SEC;
					now.tv_usec = (int)((double)(clocks % UCLOCKS_PER_SEC) * 1000000.0 / UCLOCKS_PER_SEC);
				}
				else
				{
					gettimeofday (&now, NULL);
				}
				if ((lastscroll.tv_sec | lastscroll.tv_usec) == 0)
				{
					diff = 1000000/20;
				}
				else
				{
					diff = (now.tv_sec-lastscroll.tv_sec)*1000000+now.tv_usec-lastscroll.tv_usec;
				}
#endif
				lastscroll = now;

				if (diff > 0)
				{
					// [RH] Also make scroll rate constant across different resolutions
					rate = (BCINT)(MoveSpeed * ScrMaxX * diff * 0.00000005 / Scale);
					if (rate <= 0)
					{
						rate = 1;
					}
					OrigX += moveX * rate;
					OrigY += moveY * rate;

					RedrawMap = REDRAW_QUICK;
				}
				else
				{
					RedrawMap = REDRAW_QUICKHOLD;
				}
			}
		}
		if (RedrawMap != REDRAW_QUICK && RedrawMap != REDRAW_QUICKHOLD)
		{
#ifdef __WATCOMC__
			lastscroll = 0;
#else
			lastscroll.tv_sec = lastscroll.tv_usec = 0;
#endif
		}
		GetMouseCoords( &PointerX, &PointerY, &buttons);
		if (buttons == 1 && !StretchSelBox)
		{
			count++;
			if (count >= ((StretchDelay + 3) * 50))
			{
				PlaySound(50,10);
				SelBoxX = MAPX( PointerX);
				SelBoxY = MAPY( PointerY);
				StretchSelBox = TRUE;
				GrMouseSetColors(TranslateToGameColor(YELLOW),GrNOCOLOR);
				DraggingObject = FALSE;
			}
		}
		else
		{
			count = 0;
		}
    }
}

void DrawMenuLetters(BCINT color)
{
	SetColor(color);
	DrawScreenText( 7,  4, "F    E    S      M     i       O           C");
	DrawScreenText( ScrMaxX - 50, 4, "H");
}

void DoHelp(BCINT EditMode, BCINT GridScale)
{
	BCINT x0, y0, MaxY, MaxX;

	if (InfoShown)
		MaxY = ScrMaxY;
	else
		MaxY = ScrMaxY + 13;
	MaxX = ScrMaxX;

	x0 = ScrMaxX - 140;
	y0 = ScrMaxY - 210;
	if (VideoMode < 3)
		y0 -= 91;
	if (InfoShown) 
		y0 -= 13; 
	if (EditMode == OBJ_THINGS && MouseInArea(2, MaxY - 132, 360, 116))
		DisplayHelp( EditMode, GridScale);
	else if (DispSprite && EditMode == OBJ_THINGS && MouseInArea(MaxX - 120, MaxY - 132, 120, 120)) 
		HelpThingBox();
	else if (EditMode == OBJ_LINEDEFS && MouseInArea(2, MaxY - 142, 258, 120))
		DisplayHelp( EditMode, GridScale);
	else if (DispSprite && EditMode == OBJ_LINEDEFS && MouseInArea(x0, y0, 136, 210)) 
		HelpTextureBox();
	else if (EditMode == OBJ_SECTORS && MouseInArea(2, MaxY - 92, 280, 76))
		DisplayHelp( EditMode, GridScale);
	else if (DispSprite && EditMode == OBJ_SECTORS && MouseInArea(MaxX - 72, MaxY - 148, 72, 140)) 
		HelpSectorBox();
	else if (EditMode == OBJ_LINEDEFS && MouseInArea(260, MaxY - 96, 380, 80))
		DisplayHelp( EditMode, GridScale);
	else 
		DisplayHelp( EditMode, GridScale);
}

void ShowHelpHere(BCINT color)
{
	SetColor(color);
	DrawScreenText( ScrMaxX - 50, 4, "Help");
}

void GetCount(SelPtr Selected, BCINT Start)
{
	SelPtr n;
	Count = Start;
	for(n = Selected; n; n = n->next)
		Count++;
	if (!Selected && DraggingObject && Count == 1)
		Count = 0;
}

void DrawArcSpots (int *circlex, int *circley, int circlepoints)
{
	int i, s = OBJSIZE/2/Scale;

	DrawMapLine( circlex[0] - s, circley[0] - s, circlex[0] + s, circley[0] + s);
	DrawMapLine( circlex[0] - s, circley[0] + s, circlex[0] + s, circley[0] - s);

	for (i = 0; i < circlepoints; ++i)
	{
		DrawMapHollowBox (circlex[1+i] - s, circley[1+i] - s, circlex[1+i] + s, circley[1+i] + s);
	}
}


/* end of file */
